# Copyright 2020-2021 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyjls import Writer, SourceDef, SignalDef, SignalType, DataType
import numpy as np


"""This modules writes Joulescope streaming data to JLS v2 files.

The JLS v2 file format is very flexible with multiple options.  This module
provides a simple interface to connect Joulescope streaming data to
the JLS v2 file writer.
"""


SIGNALS = {
    'current': (1, 'A'),
    'voltage': (2, 'V'),
    'power': (3, 'W'),
}


def _signals_validator(s):
    result = []
    if isinstance(s, str):
        p = s.split(',')
    else:
        p = s
    for signal_name in p:
        k = signal_name.lower()
        if k not in SIGNALS:
            raise ValueError(f'unsupported signal {signal_name}')
        result.append(k)
    return result


def _sampling_rate_validator(s):
    if isinstance(s, str):
        if s.endswith('Hz'):
            s = s[:-2]
        n, u = s.split()
        s = int(n)
        if u[0] == 'M':
            s *= 1000000
        elif u[0] == 'k':
            s *= 1000
    return int(s)


class JlsWriter:

    def __init__(self, device, filename, signals=None):
        """Create a new JLS file writer instance.

        :param device: The Joulescope device instance.
        :param filename: The output ".jls" filename.
        :param signals: The signals to record as either a list of string names
            or a comma-separated string.  The supported signals include
            ['current', 'voltage', 'power']

        This class implements joulescope.driver.StreamProcessApi and may also
        be used as a context manager.
        """
        self._device = device
        self._filename = filename
        if signals is None:
            signals = ['current', 'voltage']
        signals = _signals_validator(signals)
        self._signals = signals
        self._wr = None
        self._idx = 0

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def open(self):
        """Open and configure the JLS writer file."""
        self.close()
        info = self._device.info()
        sampling_rate = self._device.parameter_get('sampling_frequency')
        sampling_rate = _sampling_rate_validator(sampling_rate)

        source = SourceDef(
            source_id=1,
            name=str(self._device),
            vendor='Jetperch',
            model=info['ctl']['hw'].get('model', 'JS110'),
            version=info['ctl']['hw'].get('rev', '-'),
            serial_number=info['ctl']['hw']['sn_mfg'],
        )

        wr = Writer(self._filename)
        try:
            wr.source_def_from_struct(source)
            for s in self._signals:
                idx, units = SIGNALS[s]
                s_def = SignalDef(
                    signal_id=idx,
                    source_id=1,
                    signal_type=SignalType.FSR,
                    data_type=DataType.F32,
                    sample_rate=sampling_rate,
                    name=s,
                    units=units,
                )
                wr.signal_def_from_struct(s_def)

        except Exception:
            wr.close()
            raise

        self._wr = wr
        return wr

    def close(self):
        """Finalize and close the JLS file."""
        wr, self._wr = self._wr, None
        if wr is not None:
            wr.close()

    def stream_notify(self, stream_buffer):
        """Handle incoming stream data.

        :param stream_buffer: The :class:`StreamBuffer` instance which contains
            the new data from the Joulescope.
        :return: False to continue streaming.
        """
        # called from USB thead, keep fast!
        # long-running operations will cause sample drops
        start_id, end_id = stream_buffer.sample_id_range
        if self._idx < end_id:
            data = stream_buffer.samples_get(self._idx, end_id, fields=self._signals)
            for s in self._signals:
                x = np.ascontiguousarray(data['signals'][s]['value'])
                idx = SIGNALS[s][0]
                self._wr.fsr_f32(idx, self._idx, x)
            self._idx = end_id
        return False

    def fsr_f32(self, signal_name, sample_id, x):
        """Write signal data directly.

        :param signal_name: The signal name string.
        :param sample_id: The starting sample id for x.
        :param x: The 1-d ndarray of float sample data
        """
        idx = SIGNALS[signal_name][0]
        sample_id = int(sample_id)
        if x.dtype != np.float32:
            x = x.astype(np.float32)
        self._wr.fsr_f32(idx, sample_id, np.ascontiguousarray(x))



# Copyright 2019 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This executable configures Joulescope parameters.
"""

from joulescope.driver import scan
import logging


def parser_config(p):
    """Configure Joulescope parameters."""
    p.add_argument('--config',
                   choices=['auto', 'off', 'ignore'],
                   default='ignore',
                   help='The configuration defaults.')
    p.add_argument('--select',
                   help='Select a single Joulescope by serial number.')
    p.add_argument('assigns',
                   nargs='*',
                   help='A parameter name=value to assign.  See parameters_v1.py for details.')
    return on_cmd


def on_cmd(args):
    rv = 0
    devices = scan(name='Joulescope', config=args.config)
    for device in devices:
        try:
            if args.select and int(device.device_serial_number) != int(args.select):
                continue
            print(f'{device}')
            with device:
                for name_value in args.assigns:
                    try:
                        name, value = name_value.split('=')
                        device.parameter_set(name, value)
                    except Exception:
                        logging.exception(f'{device} parameter {name_value} failed')
                        print(f'{device} parameter {name_value} failed')
                pass
        except Exception:
            logging.exception(f'{device}: failed')
            print(f'{device}: failed')
            rv = 1
    return rv



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from joulescope import datafile
import binascii
from joulescope import public_keys
from joulescope.time import seconds_to_timestamp
import io
import json
import dateutil.parser
import numpy as np
import logging


log = logging.getLogger(__name__)
PRODUCT = '1.1.1'


def _stuffed(x):
    while len(x) < 9:
        x.append(x[-1])
    x[7] = 0.0
    x[8] = float('nan')
    return np.array(x, dtype=np.float32)


def _version_str_to_u32(s: str):
    major, minor, patch = [int(x) for x in s.split('.')]
    return (((major & 0xff) << 24) |
            ((minor & 0xff) << 16) |
            (patch & 0xffff))


def _np_list(x):
    return np.array(x, dtype=np.float32).tolist()


def raw_split(data):
    """Split raw data into current, voltage, i_range and missing_sample_count.

    :param data: The (N, 2) or (N x 2, ) np.ndarray with dtype np.uint16.
        Either way, the current and voltage samples are interleaved as
        they arrive from Joulescope over USB.
        The least significant 2 bits as indicators.
    :return: (current, voltage, i_range, missing_sample_count) where missing
        samples are detected by the per-sample bit toggle.  Current and Voltage
        are N np.ndarray(dtype=np.uint16) with 14-bit right-shifted values.
        i_range is N np.ndarray(dtype=np.uint8) with 3-bit right shifted values.
        missing_sample_count is a scalar integer.
    """
    if len(data.shape) == 2:
        data = data.reshape((np.prod(data.shape), ))

    # get i_range
    data_i_range_low = np.bitwise_and(data[::2], 0x0003)
    data_i_range_high = np.bitwise_and(data[1::2], 0x0001)
    data_i_range_high = np.left_shift(data_i_range_high, 2, out=data_i_range_high)
    i_range = np.bitwise_or(data_i_range_low, data_i_range_high, out=data_i_range_low)
    i_range = i_range.astype(np.uint8)

    # every sample also contains a toggling sample indicator
    # to check proper packet construction.
    # This check will not detect samples lost at packet boundaries,
    # but this is good enough to detect system problems.  The hardware
    # ensures that individual samples should not be lost, just packets.
    toggle = np.bitwise_and(data[1::2], 0x0002)
    toggle = np.right_shift(toggle, 1, out=toggle)
    missing_sample_count = len(toggle) - 1 - \
        np.count_nonzero(np.bitwise_xor(toggle[:-1], toggle[1:]))

    # get 14-bit right justified current and voltage data
    data = np.right_shift(data, 2)
    i_raw = data[::2]
    v_raw = data[1::2]
    return i_raw, v_raw, i_range, missing_sample_count


class Calibration:
    """Manage Joulescope calibration data."""

    def __init__(self):
        self.version = '1.0.0'
        self.product_id = 1
        self.vendor_id = 1
        self.subtype_id = 1
        self.time = dateutil.parser.parse('2018-01-01T00:00:00.00000Z')
        self.serial_number = '00000000000000000000000000000000'
        self.current_offset = np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, np.nan], dtype=np.float32)
        self.current_gain = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 0.0, np.nan], dtype=np.float32)
        self.voltage_offset = np.array([0.0, 0.0], dtype=np.float32)
        self.voltage_gain = np.array([1.0, 1.0], dtype=np.float32)
        self.signed = False
        self.data = None  # store load verbatim for future direct write

    def json(self):
        return {
            'version': self.version,
            'time': self.time.isoformat(),
            'timestamp': str(seconds_to_timestamp(self.time.timestamp())),
            'product_name': 'Joulescope JS110',
            'vendor_name': 'Jetperch LLC',
            'subtype_name': 'Calibration',
            'product': f'{self.product_id}.{self.vendor_id}.{self.subtype_id}',
            'serial_number': self.serial_number,
            'voltage': {
                'offset': _np_list(self.voltage_offset),
                'gain': _np_list(self.voltage_gain),
            },
            'current': {
                'offset': _np_list(self.current_offset),
                'gain': _np_list(self.current_gain),
            },
        }

    def save(self, private_key=None):
        """Save calibration to bytes.

        :param private_key: The private key used to sign the calibration.
            None (default) does not sign the calibration.
            The Joulescope software will display warnings if the calibration
            is not signed using a valid key.
        :return: The calibration as bytes.
        """
        if isinstance(self.time, str):
            self.time = dateutil.parser.parse(self.time)
        fh = io.BytesIO()
        dfw = datafile.DataFileWriter(fh)
        if private_key is not None:
            dfw.signature_start(private_key)
        dfw.append_header(
            timestamp=seconds_to_timestamp(self.time.timestamp()),
            version=_version_str_to_u32(self.version),
            product_id=self.product_id,
            vendor_id=self.vendor_id,
            subtype_id=self.subtype_id,
            hardware_compatibility=0,
            serial_number=binascii.unhexlify(self.serial_number),
        )
        self.current_offset = np.concatenate((self.current_offset[:7], [0.0, np.nan]))
        self.current_gain = np.concatenate((self.current_gain[:7], [0.0, np.nan]))
        if len(self.current_offset) != 9 or len(self.current_gain) != 9:
            raise ValueError('Invalid length for current')
        if len(self.voltage_offset) != 2 or len(self.voltage_gain) != 2:
            raise ValueError('Invalid length for voltage')
        cal = self.json()
        dfw.append(datafile.TAG_DATA_JSON, json.dumps(cal, allow_nan=True).encode('utf-8'))
        if private_key is not None:
            dfw.signature_end()
        dfw.finalize()
        return bytes(fh.getbuffer())

    def load(self, data, keys=None):
        """Load calibration from bytes.

        :param data: The bytes containing the calibration.
        :param keys: The list of allowed public keys.  None (default)
            uses the official signing key.
        :return: This instance.
        :raise ValueError: on invalid data.
        """
        self.data = data
        fh = io.BytesIO(data)
        r = datafile.DataFileReader(fh)
        self.signed = True
        tag, value = next(r)
        public_key = value[8:]
        if keys is None:
            keys = [public_keys.CALIBRATION_SIGNING, public_keys.RECALIBRATION_01_SIGNING]
        if datafile.TAG_SIGNATURE_START != tag:
            self.signed = False
            log.warning('Invalid format: missing signature start tag')
        else:
            if public_key not in keys:
                self.signed = False
                log.warning('Invalid signing key')
            tag, header = next(r)
        if datafile.TAG_HEADER != tag:
            raise ValueError('Invalid format: missing header tag')
        tag, cal = next(r)
        if tag not in [datafile.TAG_DATA_JSON, datafile.TAG_CALIBRATION_JSON]:
            raise ValueError('Invalid format: missing calibration tag')
        if self.signed:
            tag, _ = next(r)
            if datafile.TAG_SIGNATURE_END != tag:
                self.signed = False
                log.warning('Invalid format: missing signature end tag')
        cal = json.loads(cal.decode('utf-8'))
        if cal['product'] != PRODUCT:
            raise ValueError('Invalid calibration')
        self.version = cal['version']
        self.time = dateutil.parser.parse(cal['time'])
        self.serial_number = cal['serial_number']
        self.current_offset = _stuffed(cal['current']['offset'])
        self.current_gain = _stuffed(cal['current']['gain'])
        self.voltage_offset = np.array(cal['voltage']['offset'], dtype=np.float32)
        self.voltage_gain = np.array(cal['voltage']['gain'], dtype=np.float32)
        return self

    def transform(self, data, v_range=None):
        """Apply the calibration to transform raw data.

        :param data: The (N, 2) or (N x 2, ) np.ndarray with dtype np.uint16.
            Either way, the current and voltage samples are interleaved as
            they arrive from Joulescope over USB.
            The least significant 2 bits as indicators.
        :param v_range: The selected voltage ranged for the conversion.
        :return: (current, voltage, missing_sample_count) where missing samples
            is detected by the per-sample bit toggle.
        """
        v_range = 0 if v_range is None else int(v_range)
        if len(data.shape) == 2:
            data = data.reshape((np.prod(data.shape), ))

        i, v, i_range, missing_sample_count = raw_split(data)
        idx = np.logical_and(np.logical_and(i_range == 7, i == 0x3fff), v == 0x3fff)
        i_range[idx] += 1

        # get 14-bit right justified current and voltage data
        i = i.astype(np.float32)
        i += self.current_offset[i_range]
        i *= self.current_gain[i_range]
        v = v.astype(np.float32)
        v += self.voltage_offset[v_range]
        v *= self.voltage_gain[v_range]
        v[np.isnan(i)] = np.nan

        return i, v, missing_sample_count



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import signal
import time
import logging
from joulescope.driver import scan_require_one
from joulescope.data_recorder import DataRecorder


def parser_config(p):
    """Capture data from Joulescope."""
    p.add_argument('--duration',
                   type=float,
                   help='The capture duration in seconds.')
    p.add_argument('--contiguous',
                   type=float,
                   help='The contiguous capture duration (no missing samples) in seconds.')
    p.add_argument('filename',
                   help='The filename for output data.')
    p.add_argument('--profile',
                   choices=['cProfile', 'yappi'],
                   help='Profile the capture')
    return on_cmd


def on_cmd(args):
    device = scan_require_one(name='Joulescope', config='auto')
    f = lambda: run(device, filename=args.filename,
                    duration=args.duration,
                    contiguous_duration=args.contiguous)
    if args.profile is None:
        return f()
    elif args.profile == 'cProfile':
        import cProfile
        import pstats
        cProfile.runctx('f()', globals(), locals(), "Profile.prof")
        s = pstats.Stats("Profile.prof")
        s.strip_dirs().sort_stats("time").print_stats()
    elif args.profile == 'yappi':
        import yappi
        yappi.start()
        rv = f()
        yappi.get_func_stats().print_all()
        yappi.get_thread_stats().print_all()
        return rv
    else:
        raise ValueError('bad profile argument')


def run(device, filename, duration=None, contiguous_duration=None):
    quit_ = False

    def do_quit(*args, **kwargs):
        nonlocal quit_
        quit_ = 'quit from SIGINT'

    def on_stop(event, message):
        nonlocal quit_
        quit_ = 'quit from stop duration'

    recorder = None
    signal.signal(signal.SIGINT, do_quit)
    try:
        device.open()
        recorder = DataRecorder(filename,
                                calibration=device.calibration)
        device.stream_process_register(recorder)
        device.start(stop_fn=on_stop, duration=duration,
                     contiguous_duration=contiguous_duration)
        time_last = time.time()
        sample_id_last = 0
        sample_id_incr = 1000000
        sample_id_next = sample_id_last + sample_id_incr
        status_failures = 0
        while not quit_:
            time.sleep(0.01)
            time_now = time.time()
            if time_now - time_last > 1.0:
                s = device.status()
                if s.get('driver', {}).get('return_code', {}).get('value', 1):
                    status_failures += 1
                    if status_failures >= 3:
                        raise RuntimeError(f'status_failures = {status_failures}')
                logging.getLogger().info(s)
                time_last = time_now
            while device.stream_buffer.sample_id_range[-1] >= sample_id_next:
                # todo save
                sample_id_last = sample_id_next
                sample_id_next += sample_id_incr
        device.stop()
    except Exception as ex:
        logging.getLogger().exception('while capturing data')
        print('Data capture failed')
        return 1
    finally:
        if recorder is not None:
            recorder.close()
        device.close()
    print('done capturing data: %s' % quit_)
    return 0



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
from joulescope import time


class TestTimestamp(unittest.TestCase):

    def test_epoch(self):
        self.assertEqual(1514764800.0, time.EPOCH)

    def test_seconds_to_timestamp(self):
        self.assertEqual(0, time.seconds_to_timestamp(time.EPOCH))

    def test_timestamp_to_seconds(self):
        self.assertEqual(time.EPOCH, time.timestamp_to_seconds(0))

    def test_conversion(self):
        now = time.timestamp_now()
        s = time.timestamp_to_seconds(now)
        t = time.seconds_to_timestamp(s)
        self.assertEqual(now, t)



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The USB backend which must be implemented for each platform type.

This module defines the USB backend.  Each target platform
(such as Windows, Mac OS/X and Linux), must implement backend that conforms
to this API.

This API is **not** thread-safe.  All methods and functions must be invoked
from a single thread.
"""


class DeviceEvent:
    ENDPOINT_CALLBACK_STOP = -1  # a callback indicated that streaming should stop
    UNDEFINED = 0
    COMMUNICATION_ERROR = 1  # an communicate error that prevents this device from functioning, such as device removal
    ENDPOINT_CALLBACK_EXCEPTION = 2  # a callback threw an exception


class DeviceDriverApi:
    """The device driver API.

    This API is **not** thread-safe.  All methods must be invoked from a
    single thread.
    """

    def __str__(self):
        """Get the user-friendly device string.

        :return: f'{product_id_str}:{serial_number_str}'
        :raise IOError: On failure.

        WARNING: This function must correctly identify the device BEFORE it
        is opened.  Therefore, it must only use the information available
        from USB enumeration.
        """
        raise NotImplementedError()

    @property
    def serial_number(self):
        """Get the assigned serial number.

        :return: The serial number string.

        This attribute is valid even before the device is opened.
        """
        raise NotImplementedError()

    def open(self, event_callback_fn):
        """Open the USB device.

        :param event_callback_fn: The function(event, message) to call on
            asynchronous events, mostly to allow robust handling of device
            errors.  "event" is one of the :class:`DeviceEvent` values,
            and the message is a more detailed description of the event.
        :raise IOError: On failure.

        The event_callback_fn may be called asynchronous and from other
        threads.  The event_callback_fn must implement any thread safety.
        """
        raise NotImplementedError()

    def close(self):
        """Close the USB device."""
        raise NotImplementedError()

    def control_transfer_out(self, cbk_fn, recipient, type_, request, value=0, index=0, data=None) -> bool:
        """Perform a control transfer with data from host to device.

        :param cbk_fn: The function called with the class:`ControlTransferResponse` result.
            This method guarantees that cbk_fn will always be called.
            cbk_fn may be called BEFORE exiting this method call.
        :param recipient: The recipient which is one of ['device', 'interface', 'endpoint', 'other']
        :param type_: The type which is one of ['standard', 'class', 'vendor'].
        :param request: The bRequest value.
        :param value: The wValue value.
        :param index: The wIndex value.
        :param data: The optional data to transfer from host to device.
            None (default) skips the data phase.
        :return: True on pending, False on error.
        """
        raise NotImplementedError()

    def control_transfer_in(self, cbk_fn, recipient, type_, request, value, index, length) -> bool:
        """Perform a control transfer with data from device to host.

        :param cbk_fn: The function called with the class:`ControlTransferResponse` result.
            This method guarantees that cbk_fn will always be called.
            cbk_fn may be called BEFORE exiting this method call.
        :param recipient: The recipient which is one of ['device', 'interface', 'endpoint', 'other']
        :param type_: The type which is one of ['standard', 'class', 'vendor'].
        :param request: The bRequest value.
        :param value: The wValue value.
        :param index: The wIndex value.
        :param length: The maximum number of bytes to transfer from device to host.
        :return: True on pending, False on error.
        """
        raise NotImplementedError()

    def read_stream_start(self, endpoint_id, transfers, block_size, data_fn, process_fn, stop_fn):
        """Read a stream of data using non-blocking (overlapped) IO.

        :param endpoint_id: The target endpoint address.
        :param transfers: The number of overlapped transfers to use,
            each of block_size bytes.
        :param block_size: The length of each block in bytes which must be
            a multiple of the maximum packet size for the endpoint.
        :param data_fn: The function(data) to call on each block
            of data.  The data is an np.ndarray(dtype=uint8) containing
            the raw bytes received for each USB transaction.
            The length of data is normally block_size.
            Any value less than block_size is the last transfer
            in the transaction.

            When the device stops, it calls data_fn(None).  The
            device can stop "automatically" through errors or when data_fn
            returns True.  Call :meth:`read_stream_stop` to stop from
            the caller.

            This function will be called from the device's thread.  The
            data_fn must return quickly to ensure that the USB stream
            is not starved.

            In all cases, data_fn should return None or False to continue
            streaming.  data_fn can return True to stop the transmission.

            Most implementations use some form of non-blocking IO with
            multiple queue (overlapped) transactions that are pended
            early.  On stop, additional data may be read before the
            transaction fully stops.
        :param process_fn: The function process_fn() to call after all
            USB endpoints have been recently serviced and data_fn was
            called at least once.  The function should still be quick,
            but it can have more latency than data_fn.
        :param stop_fn: The function(event, message) called when this endpoint
            stops streaming data. See :class:`DeviceEvent` for allowed event
            values.

        Use :meth:`read_stream_stop` to stop.
        """
        raise NotImplementedError()

    def read_stream_stop(self, endpoint_id):
        """Stop a read stream.

        :param endpoint_id: The target endpoint address.

        When stop is complete, the data_fn provided to read_stream_start will
        be called with None.

        Use :meth:`read_stream_start` to start.
        """
        raise NotImplementedError()

    def status(self):
        """Get the current device status.

        :return: A dict containing the following structure:

            endpoints: {
                pipe_id: { name: {value: v, units: u}, ...}
                ...
            }
        """
        raise NotImplementedError()

    def signal(self):
        """Signal that an external event occurred.

        This method allows another thread to cause the wait in process
        to activate.
        """
        raise NotImplementedError()

    def process(self, timeout=None):
        """Process any pending events.

        :param timeout: The timeout in float seconds.

        This method uses the operating-system specific method to wait on
        pending events, such select and WaitForMultipleObjects.
        """
        raise NotImplementedError()


class DeviceNotify:

    def __init__(self, cbk):
        """Start device insertion/removal notification.

        :param cbk: The function called on device insertion or removal.  The
            arguments are (inserted, info).  "inserted" is True on insertion
            and False on removal.  "info" contains platform-specific details
            about the device.  In general, the application should rescan for
            relevant devices.
        """
        pass

    def close(self):
        """Close and stop the notifications."""
        raise NotImplementedError()


def scan(name: str=None):
    """Scan for attached devices.

    :param name: The case-insensitive name of the device to scan.
    :return: The list of attached backend :class:`Device` instances.
    """
    raise NotImplementedError()



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Define the JouleScope parameters available to the application.
"""

from .parameter import Parameter


def _buffer_duration_validator(x):
    x = float(x)
    if x < 0:
        raise ValueError('too small')
    return int(x + 0.5001)

# list of [param_name, permission, path, default, [(value_name, value), ...]]
# * permission is either 'r' (read-only) or 'rw' (read-write)
# * path is the USB (request, index) for the transaction
# * default is the name for the default value

PARAMETERS = [
    Parameter(
        name='sensor_power',
        brief='Force the sensor power on or off.',
        detail='Altering this parameter during operation may cause errors.',
        path='setting',
        default='on',
        options=[('off', 0), ('on', 1)],
        flags=['developer'],
    ),

    Parameter(
        name='source',
        brief='Select the streaming data source.',
        path='setting',
        default='off',
        options=[
            ('off',             0x00),
            ('raw',             0xC0, ['on']),
            ('pattern_usb',     0x09),
            ('pattern_control', 0x0A),
            ('pattern_sensor',  0xAF),
        ],
        flags=['developer'],
    ),
    Parameter(
        name='i_range',
        brief='Select the current measurement range (shunt resistor)',
        path='setting',
        default='off',
        options=[
            ('auto',   0x80, ['on']),
            ('10 A',   0x01, ['0', 0]),
            ('2 A',    0x02, ['1', 1]),
            ('180 mA', 0x04, ['2', 2]),
            ('18 mA',  0x08, ['3', 3]),
            ('1.8 mA', 0x10, ['4', 4]),
            ('180 µA', 0x20, ['5', 5]),
            ('18 µA',  0x40, ['6', 6]),
            ('off',    0x00),
        ],
    ),
    Parameter(
        name='v_range',
        brief='Select the voltage measurement range (gain)',
        path='setting',
        default='15V',
        options=[
            ('15V', 0, ['low', 0]),
            ('5V',  1, ['high', 1]),
        ],
    ),
    Parameter(
        name='ovr_to_lsb',
        brief='Map overflow flags to the LSBs',
        path='setting',
        default='off',
        options=[('off', 0), ('on', 1)],
        flags=['developer'],
    ),
    Parameter(
        name='trigger_source',
        brief='Select the trigger source',
        path='extio',
        default='auto',
        options=[
            ('auto', 0),
            ('gpi0', 2),
            ('gpi1', 3),
        ],
        flags=['hidden'],  # feature not ready yet
    ),

    # --- EXTIO ---
    Parameter(
        name='io_voltage',
        brief='The GPI/O high-level voltage.',
        path='extio',
        default='3.3V',
        options=[
            ('1.8V', 1800),
            ('2.1V', 2100),
            ('2.5V', 2500),
            ('2.7V', 2700),
            ('3.0V', 3000),
            ('3.3V', 3300),
            ('3.6V', 3600),
            ('5.0V', 5000),
        ],
    ),
    Parameter(
        name='gpo0',
        brief='The GPO bit 0 output value.',
        path='extio',
        default='0',
        options=[
            ('0', 0, [0]),
            ('1', 1, [1]),
            # ('start_pulse', 2),   # reserved, but not yet implemented
            # ('sample_toggle', 3), # reserved, but not yet implemented
        ],
    ),
    Parameter(
        name='gpo1',
        brief='The GPO bit 1 output value.',
        path='extio',
        default='0',
        options=[
            ('0', 0, [0]),
            ('1', 1, [1]),
        ],
    ),
    Parameter(
        name='current_lsb',
        brief='The current signal least-significant bit mapping.',
        path='extio',
        default='normal',
        options=[
            ('normal', 0),
            ('gpi0', 2),
            ('gpi1', 3),
        ],
    ),
    Parameter(
        name='voltage_lsb',
        brief='The voltage signal least-significant bit mapping.',
        path='extio',
        default='normal',
        options=[
            ('normal', 0),
            ('gpi0', 2),
            ('gpi1', 3),
        ],
    ),

    Parameter(
        name='control_test_mode',
        brief='Set the test mode',
        path='setting',
        default='normal',
        options=[
            ('normal', 0x03),
            ('usb',    0x81),
            ('fpga',   0x82),
            ('both',   0x83),  # also set 'source' to 'pattern_sensor'
        ],
        flags=['developer'],
    ),
    Parameter(
        name='transfer_length',
        brief='Set the USB transfer length in packets',
        path='setting',
        default='256',
        options=[('%d' % (2**x), (2**x)) for x in range(0, 9)],
        units='packets',
        flags=['developer', 'skip_update'],
    ),
    Parameter(
        name='transfer_outstanding',
        brief='Set the maximum number of USB transfers issued simultaneously',
        path='setting',
        default='8',
        options=[('%d' % (2**x), (2**x)) for x in range(0, 4)],
        flags=['developer', 'skip_update']
    ),

    Parameter(
        name='current_ranging',  # virtual parameter!
        path='current_ranging',
        default=None,
        options=None,  # parse into current_ranging fields: type, samples_pre, samples_window, samples_post
        flags=['hidden'],
    ),
    Parameter(
        name='current_ranging_type',
        brief='The filter type.',
        path='current_ranging',
        default='interp',
        options=[
            ('off', 'off'),
            ('mean', 'mean'),
            ('interp', 'interp', ['interpolate']),
            ('NaN', 'nan', ['nan'])
        ],
    ),
    Parameter(
        name='current_ranging_samples_pre',
        brief='The number of samples before the range switch to include.',
        detail='Only valid for type "mean" - ignored for "off", "interp", and "NaN".',
        path='current_ranging',
        default='1',
        options=[(str(d), d, [d]) for d in range(9)]
    ),
    Parameter(
        name='current_ranging_samples_window',
        brief='The number of samples to adjust.',
        detail='Use "n" for automatic duration based upon known response time. ' +
               'Use "m" for shorter automatic duration that may result in min/max distortion.',
        path='current_ranging',
        default='n',
        options=[('m', 'm'), ('n', 'n')] + [(str(d), d, [d]) for d in range(13)]
    ),
    Parameter(
        name='current_ranging_samples_post',
        brief='The number of samples after the range switch to include.',
        detail='Only valid for type "mean" - ignored for "off", "interp", and "NaN".',
        path='current_ranging',
        default='1',
        options=[(str(d), d, [d]) for d in range(9)]
    ),

    Parameter(
        name='buffer_duration',
        brief='The amount of sample data to store in memory.',
        detail='Use care when setting this value. ' +
               'The software requires 1.5 GB of RAM for every 60 seconds at ' +
               'the full 2 MSPS rate.',
        path='setting',
        default='30 seconds',
        options=[
            ('15 seconds', 15, [15]),
            ('30 seconds', 30, [30]),
            ('1 minute', 60, [60]),
            ('2 minutes', 120, [120]),
            ('5 minutes', 300, [300]),
            ('10 minutes', 600, [600]),
            ('20 minutes', 60*20, [60*20]),
            ('1 hour', 3600, [3600]),
            ('2 hours', 2*3600, [2*3600]),
            ('5 hours', 5*3600, [5*3600]),
            ('10 hours', 10*3600, [10*3600]),
            ('1 day', 24 * 3600, [24*3600])],
        units='seconds',
        validator=_buffer_duration_validator,
        flags=['skip_update']
    ),
    Parameter(
        name='reduction_frequency',
        brief='The rate that the device produces statistics, including multimeter values.',
        path='setting',
        default='2 Hz',
        options=[
            ('100 Hz', 100, [100]),
            ('50 Hz', 50, [50]),
            ('20 Hz', 20, [20]),
            ('10 Hz', 10, [10]),
            ('5 Hz', 5, [5]),
            ('2 Hz', 2, [2]),
            ('1 Hz', 1, [1])],
        units='Hz',
        flags=['skip_update']
    ),
    Parameter(
        name='sampling_frequency',
        brief='The rate that the device produces samples.',
        path='setting',
        default='2 MHz',
        options=[
            ('2 MHz', 2000000, [2000000, 'auto', None, 'default']),
            ('1 MHz', 1000000, [1000000]),
            ('500 kHz', 500000, [500000]),
            ('200 kHz', 200000, [200000]),
            ('100 kHz', 100000, [100000]),
            ('50 kHz', 50000, [50000]),
            ('20 kHz', 20000, [20000]),
            ('10 kHz', 10000, [10000]),
            ('5 kHz', 5000, [5000]),
            ('2 kHz', 2000, [2000]),
            ('1 kHz', 1000, [1000]),
            ('500 Hz', 500, [500]),
            ('200 Hz', 200, [200]),
            ('100 Hz', 100, [100]),
            ('50 Hz', 50, [50]),
            ('20 Hz', 20, [20]),
            ('10 Hz', 10, [10])],
        units='Hz',
        flags=['skip_update']
    ),

    Parameter(name='model', path='info', flags=['read_only', 'hidden']),
    Parameter(name='device_serial_number', path='info', flags=['read_only', 'hidden']),
    Parameter(name='hardware_serial_number', path='info', flags=['read_only', 'hidden']),
]


def _lookup_construct(x):
    fwd = {}
    rev = {}
    for p in x:
        d_fwd = {}
        d_rev = {}
        fwd[p.name] = d_fwd
        rev[p.name] = d_rev
        for value_name, value, aliases in p.options:
            d_fwd[value_name] = value
            d_rev[value] = value_name
            for alias in aliases:
                d_fwd[alias] = value
    return fwd, rev


PARAMETERS_DICT = dict((p.name, p) for p in PARAMETERS)

_TO_VALUE, _TO_NAME = _lookup_construct(PARAMETERS)


def name_to_value(param_name, value_name):
    return _TO_VALUE[param_name][value_name]


def value_to_name(param_name, value):
    return _TO_NAME[param_name][value]


PARAMETERS_DEFAULTS = {
    'auto': {
        'i_range': 'auto',
        'v_range': '15V',
        'source': 'on',
    },
    'off': {
        'source': 'off',
    },
    'ignore': {},
}



# Copyright 2019 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import tempfile
import os
import sys
from joulescope.file_replace import FileReplace


class TestFileReplace(unittest.TestCase):

    def setUp(self):
        f = tempfile.NamedTemporaryFile(delete=False)
        f.close()
        self._path = f.name

    def tearDown(self) -> None:
        if os.path.isfile(self._path):
            os.remove(self._path)

    def confirm(self, contents):
        with open(self._path, 'rt', encoding='utf-8') as f:
            d = f.read()
        self.assertEqual(contents, d)

    def test_does_not_exist(self):
        os.remove(self._path)
        with FileReplace(self._path, mode='wt') as f:
            f.write('hello')
        self.confirm('hello')

    def test_file_exists(self):
        with FileReplace(self._path, mode='wt') as f:
            f.write('hello')
        self.confirm('hello')

    def test_file_delete_during_replace(self):
        with FileReplace(self._path, mode='wt') as f:
            f.write('hello')
            os.remove(self._path)
        self.confirm('hello')

    def test_exception_during_replace(self):
        with open(self._path, 'wt', encoding='utf-8') as f:
            f.write('original')
        with self.assertRaises(RuntimeError):
            with FileReplace(self._path, mode='wt') as f:
                f.write('hello')
                raise RuntimeError('doh')
        self.confirm('original')

    def test_file_open(self):
        if sys.platform.startswith('win'):
            # windows has file locking by default
            with self.assertRaises(PermissionError):
                with open(self._path, 'wt', encoding='utf-8') as f1:
                    f1.write('original')
                    with FileReplace(self._path, mode='wt') as f2:
                        f2.write('hello')
            self.confirm('original')




# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test spans
"""

import unittest
from joulescope.span import Span
import numpy as np


class TestSpan(unittest.TestCase):

    def test_trivial(self):
        s = Span([0.0, 20.0], 1.0, 11)
        self.assertEqual(20.0, s.s_limit_max)
        self.assertEqual(10.0, s.a_limit_min)
        self.assertEqual([1.0, 11.0], s.conform([1.0, 11.0]))
        self.assertEqual(1.0, s.quants_per([1.0, 11.0]))
        self.assertEqual(([2.0, 12.0], 1.0), s.conform_quant_per([1.0, 11.0]))
        span, steps_per, axis = s.conform_discrete([1.0, 11.0])
        self.assertEqual([2.0, 12.0], span)
        self.assertEqual(1.0, steps_per)
        self.assertEqual([1.0, 11.0], s.scale([1.0, 11.0]))

    def test_conform_discrete_range_trim(self):
        s = Span([10.0, 30.0], 1.0, 11)
        sc, spans_per, x = s.conform_discrete([0.0, 40.0])
        np.testing.assert_allclose([10.0, 30.0], sc)
        self.assertEqual(2, spans_per)
        np.testing.assert_allclose(np.arange(10, 31, 2, dtype=np.float64), x)

    def test_conform_discrete_min_gain(self):
        s = Span([10.0, 30.0], 1.0, 11)
        sc, spans_per, x = s.conform_discrete([19.5, 20.5])
        np.testing.assert_allclose([16, 26], sc)
        self.assertEqual(1, spans_per)
        np.testing.assert_allclose(np.arange(16, 27, dtype=np.float64), x)

    def test_conform_discrete_small(self):
        s = Span([10.0, 30.0], 1.0, 11)
        sc, spans_per, x = s.conform_discrete([19.0, 21.0])
        np.testing.assert_allclose([16.0, 26.0], sc)

    def test_conform_discrete_pivot(self):
        s = Span([10.0, 30.0], 0.001, 1000)
        span = [27.000, 27.999]
        pivot = 27.1
        sc, spans_per, x = s.conform_discrete([27.000, 27.999])
        np.testing.assert_allclose(span, sc)
        sc, spans_per, x = s.conform_discrete(span=span, gain=0.5, pivot=pivot)
        k0 = (pivot - span[0]) / (span[1] - span[0])
        k1 = (pivot - sc[0]) / (sc[1] - sc[0])
        np.testing.assert_allclose(k1, k0, rtol=0.015)

    def test_maximum_length(self):
        s = Span([10.0, 20.0], 0.01, 1000)
        self.assertEqual(1000, s.length)
        s.length = 101
        self.assertEqual(101, s.length)
        s.length = 10000
        self.assertEqual(1001, s.length)



#!/usr/bin/env python3
# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Joulescope command-line utility."""


import os
import sys
import argparse
import logging
import traceback
from joulescope.entry_points import bootloader_go, capture, capture_usb, \
    gpo_demo, info, parameter_set, program, recording, statistics, stream_test


entry_points = [bootloader_go, capture, capture_usb, gpo_demo, info,
                parameter_set, program, recording, statistics, stream_test]
"""This list of available command modules.  Each module must contain a 
parser_config(subparser) function.  The function must return the callable(args)
that will be executed for the command."""


_LOG_LEVELS = {
    'OFF': 100,
    'CRITICAL': logging.CRITICAL,
    'ERROR': logging.ERROR,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
    'ALL': 0,
}

_EPILOG = f"""\
Set the JOULESCOPE_LOG_LEVEL environment variable to change the logging level.
Options are [{', '.join(_LOG_LEVELS.keys())}].
The default is WARNING.
"""


try:
    _ui_error_msg = None
    from joulescope_ui.entry_points import ui as ui_entry_point
    entry_points.insert(0, ui_entry_point)
except ImportError:
    if len(sys.argv) == 1 or sys.argv[1] == 'ui':
        print('')
        traceback.print_exc()
        print('')
    ui_entry_point = None


def get_parser():
    parser = argparse.ArgumentParser(
        description='Joulescope™ command line tools.',
        epilog=_EPILOG,
    )
    subparsers = parser.add_subparsers(
        dest='subparser_name',
        help='The command to execute')

    for entry_point in entry_points:
        default_name = entry_point.__name__.split('.')[-1]
        name = getattr(entry_point, 'NAME', default_name)
        cfg_fn = entry_point.parser_config
        p = subparsers.add_parser(name, help=cfg_fn.__doc__)
        cmd_fn = cfg_fn(p)
        if not callable(cmd_fn):
            raise ValueError(f'Invalid command function for {name}')
        p.set_defaults(func=cmd_fn)

    subparsers.add_parser('help', help='Display the command help. Use [command] --help to display help for a specific command.')

    return parser


def run():
    log_level = os.environ.get('JOULESCOPE_LOG_LEVEL', 'WARNING').upper()
    log_level = _LOG_LEVELS.get(log_level, logging.WARNING)
    logging.basicConfig(level=log_level,
                        format="%(levelname)s:%(asctime)s:%(filename)s:%(lineno)d:%(name)s:%(message)s")
    parser = get_parser()
    args = parser.parse_args()
    if args.subparser_name is None:
        if ui_entry_point is not None:
            return ui_entry_point.run()
        else:
            print('No command provided and user interface not found.  Please specify a command.')
            parser.print_help()
            parser.exit()
    elif args.subparser_name.lower() in ['help']:
        parser.print_help()
        parser.exit()
    return args.func(args)


if __name__ == "__main__":
    sys.exit(run())



# Copyright 2018-2019 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from joulescope import usb
from joulescope.usb.device_thread import DeviceThread
from .parameters_v1 import PARAMETERS, PARAMETERS_DICT, PARAMETERS_DEFAULTS, name_to_value, value_to_name
from . import datafile
from . import bootloader
from joulescope.stream_buffer import StreamBuffer, DownsamplingStreamBuffer
from joulescope.calibration import Calibration
from joulescope.view import View
import struct
import copy
import time
import json
import io
import queue
import numpy as np
import binascii
from typing import List
import logging

log = logging.getLogger(__name__)

STATUS_REQUEST_LENGTH = 128
EXTIO_REQUEST_LENGTH = 128
SERIAL_NUMBER_LENGTH = 16
HOST_API_VERSION = 1
CALIBRATION_SIZE_MAX = 0x8000
PACKET_VERSION = 1
USB_RECONNECT_TIMEOUT_SECONDS = 15.0
SENSOR_COMMAND_TIMEOUT_SECONDS = 3.0


def _ioerror_on_bad_result(rv):
    if 0 != rv.result:
        raise IOError('usb.Device %s' % (rv,))


class UsbdRequest:

    LOOPBACK_WVALUE = 1
    """USB testing for 16-bit value, use wValue"""

    LOOPBACK_BUFFER = 2
    """USB testing for messages up to USBD_APP_LOOPBACK_BUFFER_SIZE bytes"""

    SETTINGS = 3
    """"Configure operation, including starting streaming"""

    STATUS = 4
    """Get current status (GET only)"""

    SENSOR_BOOTLOADER = 5
    """Sensor bootloader operations"""

    CONTROLLER_BOOTLOADER = 6
    """Request reboot into the controller bootloader"""

    SERIAL_NUMBER = 7
    """Request the 16-bit unique serial number."""

    CALIBRATION = 8
    """Request the calibration. wIndex 0=factory, 1=active."""

    EXTIO = 9
    """Get/set the external GPI/O settings."""

    INFO = 10
    """Get the current device information metadata JSON string."""

    TEST_MODE = 11
    """Enter a test mode."""


class SensorBootloader:
    START = 1
    RESUME = 2
    ERASE = 3
    WRITE = 4


class PacketType:
    SETTINGS = 1
    STATUS = 2
    EXTIO = 3
    INFO = 4


CALIBRATION_INDEX_TO_NAME = {
    0: 'factory',
    1: 'active',
}


CALIBRATION_INDEX = {
    None: 1,
    1: 1,
    'active': 1,
    0: 0,
    'factory': 0,
}


LOOPBACK_BUFFER_SIZE = 132
"""The maximum size of the hardware control loopback buffer for testing."""

SAMPLING_FREQUENCY = 2000000  # samples per second (Hz)


class StreamProcessApi:
    """This API is used to chain multiple processing callbacks that are called
    when new data is received from the Joulescope.

    The Joulescope driver will call notify() when new data is available.  It
    will call close() when streaming is stopped.
    """

    def start(self, stream_buffer):
        """Start a new streaming session. [optional]

        :param stream_buffer: The :class:`StreamBuffer` instance which contains
            the new data from the Joulescope.  This same instance will be passed
            to future :meth:`stream_notify` calls, too.

        This method will be called from the USB thread.  The processing
        duration must be very short to prevent dropping samples.  You
        should post any significant processing to a separate thread.
        """
        raise NotImplementedError()

    def stop(self):
        """Stop the existing streaming session. [optional]

        This method will be called from the USB thread.
        """
        raise NotImplementedError()

    def stream_notify(self, stream_buffer):
        """Notify that new data is available from the Joulescope.

        :param stream_buffer: The :class:`StreamBuffer` instance which contains
            the new data from the Joulescope.
        :return: False to continue streaming.  True to stop streaming.

        This method will be called from the USB thread.  The processing
        duration must be very short to prevent dropping samples.  You
        should post any significant processing to a separate thread.
        """
        raise NotImplementedError()

    def close(self):
        """Close the processing instance. [optional]

        This method may be called from the USB thread.
        """
        raise NotImplementedError()


def _reduction_frequency_to_reductions(frequency):
    if frequency < 1 or frequency > 100:
        raise ValueError(f'reduction_frequency {frequency} not between 1 Hz and 100 Hz')
    elif frequency < 50:
        return [200, 100, int(100 // frequency)]  # three reduction levels
    else:
        return [200, int((200 * 50) // frequency)]  # two reduction levels


def _statistics_unpack(pdu_section):
    values = struct.unpack('<qqqqiiiiiiiiiiii', pdu_section)
    samples_total, power_mean, charge, energy, samples_this, \
        samples_per_update, samples_per_second, \
        current_mean, current_min, current_max, \
        voltage_mean, voltage_min, voltage_max, \
        power_min, power_max, _ = values

    if not samples_this:
        return None

    s_start = (samples_total - samples_this)
    s_end = samples_total
    t_start = s_start / samples_per_second
    t_end = s_end / samples_per_second
    t_delta = t_end - t_start
    nan = float('nan')

    current_mean /= (1 << 27)
    current_min /= (1 << 27)
    current_max /= (1 << 27)
    voltage_mean /= (1 << 17)
    voltage_min /= (1 << 17)
    voltage_max /= (1 << 17)
    power_mean /= (1 << 34)
    power_min /= (1 << 21)
    power_max /= (1 << 21)
    charge /= (1 << 27)
    energy /= (1 << 27)

    s = {
        'time': {
            'range': {'value': [t_start, t_end], 'units': 's'},
            'delta': {'value': t_delta, 'units': 's'},
            'samples': {'value': samples_this, 'units': 'samples'},
        },
        'signals': {
            'current': {
                'µ': {'value': current_mean, 'units': 'A'},
                'σ2': {'value': nan, 'units': 'A'},
                'min': {'value': current_min, 'units': 'A'},
                'max': {'value': current_max, 'units': 'A'},
                'p2p': {'value': current_max - current_min, 'units': 'A'},
                '∫': {'value': current_mean * t_delta, 'units': 'C'},
            },
            'voltage': {
                'µ': {'value': voltage_mean, 'units': 'V'},
                'σ2': {'value': nan, 'units': 'V'},
                'min': {'value': voltage_min, 'units': 'V'},
                'max': {'value': voltage_max, 'units': 'V'},
                'p2p': {'value': voltage_max - voltage_min, 'units': 'V'},
            },
            'power': {
                'µ': {'value': power_mean, 'units': 'W'},
                'σ2': {'value': nan, 'units': 'W'},
                'min': {'value': power_min, 'units': 'W'},
                'max': {'value': power_max, 'units': 'W'},
                'p2p': {'value': power_max - power_min, 'units': 'W'},
                '∫': {'value': power_mean * t_delta, 'units': 'J'},
            },
        },
        'accumulators': {
            'charge': {
                'value': charge,
                'units': 'C',
            },
            'energy': {
                'value': energy,
                'units': 'J',
            },
        },
        'source': 'sensor',
    }
    return s


class Device:
    """The device implementation for use by applications.

    :param usb_device: The backend USB :class:`usb.device` instance.
    :param config: The initial default configuration following device open.
        Choices are ['auto', 'off', 'ignore', None].
        * 'auto': enable the sensor and start collecting data with
        current sensor autoranging.
        * 'ignore' or None: Leave the device in its existing state.
        * 'off': Turn the sensor off and disable data collection.
    """

    def __init__(self, usb_device, config=None):
        self._usb = DeviceThread(usb_device)
        self._config = config
        self._parameters = {}
        self._input_sampling_frequency = SAMPLING_FREQUENCY
        self._output_sampling_frequency = SAMPLING_FREQUENCY  # cached on open
        self._statistics_callbacks = {'stream_buffer': [], 'sensor': []}
        self._statistics_offsets = {'stream_buffer': [], 'sensor': []}
        self.stream_buffer = None
        self._streaming = False
        self._stop_fn = None
        self._process_objs = []  #: list of :class:`StreamProcessApi` compatible instances
        self._process_objs_add = []  #: list of :class:`StreamProcessApi` compatible instances
        self.calibration = None
        self._parameters_defaults = PARAMETERS_DEFAULTS
        for p in PARAMETERS:
            if 'read_only' not in p.flags and p.default is not None:
                self._parameters[p.name] = name_to_value(p.name, p.default)

    def __str__(self):
        return str(self._usb)

    @property
    def usb_device(self):
        """Get the USB backend device implementation.

        This method should only be used for unit and system tests.  Production
        code should *NEVER* access the underlying USB device directly.
        """
        return self._usb

    @property
    def input_sampling_frequency(self):
        """The original input sampling frequency."""
        return self._input_sampling_frequency

    @property
    def output_sampling_frequency(self):
        """The output sampling frequency."""
        return self.parameter_get('sampling_frequency', dtype='actual')

    @property
    def sampling_frequency(self):
        """The output sampling frequency."""
        return self.parameter_get('sampling_frequency', dtype='actual')

    @property
    def statistics_callback(self):
        """Get the registered statistics callback."""
        cbks = self._statistics_callbacks['stream_buffer']
        if len(cbks):
            return cbks[0]
        else:
            return None

    @statistics_callback.setter
    def statistics_callback(self, cbk):
        """Set the statistics callback.

        :param cbk: The callable(data) where data is a statistics data
            structure.  See the `statistics documentation <statistics.html>`_
            for details on the data format.
            This function will be called from the USB processing thread.
            Any calls back into self MUST BE resynchronized.
        """
        self._statistics_callbacks['stream_buffer'] = [cbk]

    def statistics_callback_register(self, cbk, source=None):
        """Register a statistics callback.

        :param cbk: The callable(data) where data is a statistics data
            structure.  See the `statistics documentation <statistics.html>`_
            for details on the data format.
            This function will be called from the USB processing thread.
            Any calls back into self MUST BE resynchronized.
        :param source: The statistics source where the computation is performed.
            which is one of:
            * stream_buffer: The host-side stream buffer (default).
            * sensor: The device-side FPGA.

        WARNING: calling :meth:`statistics_callback` after calling this method
        may result in unusual behavior.  Do not mix these API calls.
        """
        source = 'stream_buffer' if source is None else source
        if source not in self._statistics_callbacks:
            raise ValueError(f'Invalid source: {source}')
        self._statistics_callbacks[source].append(cbk)

    def statistics_callback_unregister(self, cbk, source=None):
        """Unregister a statistics callback.

        :param cbk: The callback previously provided to
            :meth:`statistics_callback_register`.
        :param source: The callback source.
        """
        source = 'stream_buffer' if source is None else source
        if source not in self._statistics_callbacks:
            raise ValueError(f'Invalid source: {source}')
        self._statistics_callbacks[source].remove(cbk)

    def _statistics_callback_handler(self, s):
        if s is None:
            return
        source = s['source']
        offsets = self._statistics_offsets.get(source, [])
        if not len(offsets):
            duration = s['time']['range']['value'][0]
            charge = s['accumulators']['charge']['value']
            energy = s['accumulators']['energy']['value']
            offsets = [duration, charge, energy]
            self._statistics_offsets[source] = offsets
            return
        duration_offset, charge_offset, energy_offset = offsets
        s['time']['range']['value'] = [x - duration_offset for x in s['time']['range']['value']]
        s['accumulators']['charge']['value'] -= charge_offset
        s['accumulators']['energy']['value'] -= energy_offset
        cbks = self._statistics_callbacks.get(source, [])
        for cbk in cbks:
            cbk(s)

    def statistics_accumulators_clear(self):
        """Clear the charge and energy accumulators."""
        for offsets in self._statistics_offsets.values():
            offsets.clear()

    def view_factory(self):
        """Construct a new View into the device's data.

        :return: A View-compatible instance.
        """
        view = View(self.stream_buffer, self.calibration)
        view.on_close = lambda: self.stream_process_unregister(view)
        self.stream_process_register(view)
        return view

    def parameters(self, name=None):
        """Get the list of :class:`joulescope.parameter.Parameter` instances.

        :param name: The optional name of the parameter to retrieve.
            None (default) returns a list of all parameters.
        :return: The list of all parameters.  If name is provided, then just
            return that single parameters.
        """
        if name is not None:
            for p in PARAMETERS:
                if p.name == name:
                    return copy.deepcopy(p)
            return None
        return copy.deepcopy(PARAMETERS)

    def parameter_set(self, name, value):
        """Set a parameter value.

        :param name: The parameter name
        :param value: The new parameter value
        :raise KeyError: if name not found.
        :raise ValueError: if value is not allowed
        """
        p = PARAMETERS_DICT[name]
        if name == 'current_ranging':
            self._current_ranging_split(value)
            return
        if 'read_only' in p.flags:
            log.warning('Attempting to set read_only parameter %s', name)
            return
        try:
            value = name_to_value(name, value)
        except KeyError:
            if p.validator is None:
                raise KeyError(f'value {value} not allowed for parameter {name}')
            else:
                value = p.validator(value)
        self._parameters[name] = value
        if p.path == 'setting':
            if 'skip_update' not in p.flags:
                self._stream_settings_send()
        elif p.path == 'extio':
            self._extio_set()
        elif p.path == 'current_ranging':
            self._current_ranging_set()

    def parameter_get(self, name, dtype=None):
        """Get a parameter value.

        :param name: The parameter name.
        :param dtype: The data type for the parameter.  None (default)
            attempts to convert the value to the enum string.
            'actual' returns the value in its actual type used by the driver.
        :raise KeyError: if name not found.
        """
        if name == 'current_ranging':
            pnames = ['type', 'samples_pre', 'samples_window', 'samples_post']
            values = [str(self.parameter_get('current_ranging_' + p)) for p in pnames]
            return '_'.join(values)
        p = PARAMETERS_DICT[name]
        if p.path == 'info':
            return self._parameter_get_info(name)
        value = self._parameters[name]
        if dtype == 'actual':
            return value
        try:
            return value_to_name(name, value)
        except KeyError:
            return value

    def _parameter_get_info(self, name):
        if name == 'model':
            return 'JS110'
        elif name == 'device_serial_number':
            if self.stream_buffer is None or self.calibration is None:
                return self.device_serial_number
            return self.calibration.serial_number
        elif name == 'hardware_serial_number':
            if self.stream_buffer is None:
                return None
            return self.serial_number

    @property
    def serial_number(self):
        """Get the unique 16-byte LPC54608 microcontroller serial number.

        :return: The microcontroller serial number.

        The serial number assigned during manufacturing is available using
        self.info()['ctl']['hw']['sn_mfg'].
        """
        rv = self._usb.control_transfer_in(
            'device', 'vendor',
            request=UsbdRequest.SERIAL_NUMBER,
            value=0, index=0, length=SERIAL_NUMBER_LENGTH)
        if 0 != rv.result:
            log.warning('usb control transfer failed %d', rv.result)
            return {}
        sn = bytes(rv.data)
        serial_number = binascii.hexlify(sn).decode('utf-8')
        log.info('serial number = %s', serial_number)
        return serial_number

    @property
    def device_serial_number(self):
        """Get the Joulescope device serial number assigned during manufacturing.

        :return: The serial number string.
        """
        if self._usb is None:
            return None
        return self._usb.serial_number

    def _stream_buffer_open(self):
        if self.stream_buffer:
            buffer, self.stream_buffer = self.stream_buffer, None
            del buffer
        stream_buffer_duration = self._parameters['buffer_duration']
        reduction_frequency = self._parameters['reduction_frequency']
        self._output_sampling_frequency = self.parameter_get('sampling_frequency', dtype='actual')
        reductions = _reduction_frequency_to_reductions(reduction_frequency)
        if self._input_sampling_frequency == self._output_sampling_frequency:
            log.info('Create StreamBuffer')
            self.stream_buffer = StreamBuffer(stream_buffer_duration, reductions,
                                              self._input_sampling_frequency)
        else:
            log.info('Create DownsamplingStreamBuffer')
            self.stream_buffer = DownsamplingStreamBuffer(stream_buffer_duration, reductions,
                                                          self._input_sampling_frequency,
                                                          self._output_sampling_frequency)
        self.stream_buffer.callback = self._statistics_callback_handler

    def open(self, event_callback_fn=None):
        """Open the device for use.

        :param event_callback_fn: The function(event, message) to call on
            asynchronous events, mostly to allow robust handling of device
            errors.  "event" is one of the :class:`DeviceEvent` values,
            and the message is a more detailed description of the event.
        :raise IOError: on failure.

        The event_callback_fn may be called asynchronous and from other
        threads.  The event_callback_fn must implement any thread safety.
        """
        if self.stream_buffer:
            self.close()
        self.statistics_accumulators_clear()
        self._usb.open(event_callback_fn)
        self._stream_buffer_open()
        self._current_ranging_set()
        try:
            info = self.info()
            if info is not None:
                log.info('info:\n%s', json.dumps(info, indent=2))
        except Exception:
            log.warning('could not fetch info record')
        try:
            self.calibration = self._calibration_read()
        except Exception:
            log.warning('could not fetch active calibration')
        try:
            cfg = self._parameters_defaults.get(self._config, {})
            for key, value in cfg.items():
                self.parameter_set(key, value)
        except Exception:
            log.warning('could not set defaults')
        return self

    def info(self):
        """Get the device information structure.

        :return: The device information structure.

        First implemented in 0.3.  Older firmware returns None.
        """
        rv = self._usb.control_transfer_in(
            'device', 'vendor',
            request=UsbdRequest.INFO,
            value=0, index=0, length=1024)
        if 0 != rv.result:  # firmware prior to 0.3
            return None
        if rv.data[0] != b'{'[0]:  # has header (firmware prior to 1.1.0)
            if len(rv.data) < 8:
                log.warning('info record too short')
                return None
            version, hdr_length, pdu_type = struct.unpack('<BBB', rv.data[:3])
            if version != HOST_API_VERSION:
                log.warning('info msg API version mismatch: %d != %d' % (version, HOST_API_VERSION))
                return None
            if pdu_type != PacketType.INFO:
                log.warning('info msg pdu_type mismatch: %d != %d' % (pdu_type, PacketType.INFO))
                return None
            if hdr_length != len(rv.data):
                log.warning('info msg length mismatch: %d != %d' % (hdr_length, len(rv.data)))
                return None
            json_bytes = rv.data[8:]
        else:  # just JSON string
            json_bytes = rv.data
        try:
            return json.loads(json_bytes.decode('utf-8'))
        except UnicodeDecodeError:
            log.exception('INFO has invalid unicode: %s', binascii.hexlify(rv.data[8:]))
        except json.decoder.JSONDecodeError:
            log.exception('Could not decode INFO: %s', rv.data[8:])
        return None

    def _calibration_read_raw(self, cal_idx):
        rv = self._usb.control_transfer_in(
            'device', 'vendor',
            request=UsbdRequest.CALIBRATION,
            value=cal_idx, index=0, length=datafile.HEADER_SIZE)
        if 0 != rv.result:
            log.warning('calibration_read transfer failed %d', rv.result)
            return None
        try:
            length, _ = datafile.validate_file_header(bytes(rv.data))
        except Exception:
            log.warning('invalid calibration file')
            log.info('calibration = %s', binascii.hexlify(rv.data))
            return None

        calibration = b''
        offset = 0
        while offset < length:
            # note: can only transfer 4096 (0x1000) bytes in one transfer
            # https://docs.microsoft.com/en-us/windows/desktop/api/winusb/nf-winusb-winusb_controltransfer
            k = 4096
            if k > length:
                k = length
            rv = self._usb.control_transfer_in(
                'device', 'vendor',
                request=UsbdRequest.CALIBRATION,
                value=cal_idx, index=offset, length=k)
            if 0 != rv.result:
                log.warning('calibration_read transfer failed %d', rv.result)
                return None
            chunk = bytes(rv.data)
            offset += len(chunk)
            calibration += chunk
        return calibration

    def _calibration_read(self, calibration=None) -> Calibration:
        cal_idx = CALIBRATION_INDEX[calibration]
        cal = Calibration()
        serial_number = self.serial_number
        cal.serial_number = serial_number
        try:
            cal_data = self._calibration_read_raw(cal_idx)
            if cal_data is None:
                log.info('no calibration present')
            else:
                cal.load(cal_data)
        except (ValueError, IOError):
            log.info('failed reading %s calibration', CALIBRATION_INDEX_TO_NAME[cal_idx])
        if cal.serial_number != serial_number:
            log.info('calibration serial number mismatch')
            return None
        self.calibration = cal
        self.stream_buffer.calibration_set(cal.current_offset, cal.current_gain, cal.voltage_offset, cal.voltage_gain)
        return cal

    def close(self):
        """Close the device and release resources"""
        if self.stream_buffer is not None:
            try:
                self.stop()
            except Exception:
                log.exception('USB stop failed')
            try:
                self._usb.close()
            except Exception:
                log.exception('USB close failed')
            self._stream_process_call('close')
            self.stream_buffer.callback = None
            self.stream_buffer = None

    def _wait_for_sensor_command(self, timeout=None):
        timeout = SENSOR_COMMAND_TIMEOUT_SECONDS if timeout is None else float(timeout)
        time_start = time.time()
        while True:
            dt = time.time() - time_start
            if dt > timeout:
                raise RuntimeError('timed out')
            s = self._status()
            if 0 != s['return_code']['value']:
                log.warning('Error while getting status: %s', s['return_code']['str'])
                time.sleep(0.4)
                continue
            rv = s.get('settings_result', {}).get('value', -1)
            if rv in [-1, 19]:
                time.sleep(0.010)
                continue
            return rv

    def _stream_settings_send(self):
        length = 16
        if self._streaming:
            streaming = self._parameters['control_test_mode']
        else:
            streaming = 0
        options = (self._parameters['v_range'] << 1) | self._parameters['ovr_to_lsb']
        msg = struct.pack('<BBBBIBBBBBBBB',
                          PACKET_VERSION,
                          length,
                          PacketType.SETTINGS,
                          0,  # rsvl (1 byte)
                          0,  # rsv4 (4 byte)
                          self._parameters['sensor_power'],
                          self._parameters['i_range'],  # select
                          self._parameters['source'],
                          options,
                          streaming,
                          0,  # rsv1_u8
                          0,  # rsv2_u8
                          0  # rsv3_u8
                          )
        rv = self._usb.control_transfer_out(
            'device', 'vendor', request=UsbdRequest.SETTINGS,
            value=0, index=0, data=msg)
        _ioerror_on_bad_result(rv)
        if streaming == 0:
            self._wait_for_sensor_command()

    def _extio_set(self):
        msg = struct.pack('<BBBBIBBBBBBBBII',
                          PACKET_VERSION,
                          24,
                          PacketType.EXTIO,
                          0,  # hdr_rsv1
                          0,  # hdr_rsv4
                          0,  # flags
                          self._parameters['trigger_source'],
                          self._parameters['current_lsb'],
                          self._parameters['voltage_lsb'],
                          self._parameters['gpo0'],
                          self._parameters['gpo1'],
                          0,  # uart_tx mapping reserved
                          0,  # rsv1_u8
                          0,  # rsv3_u32, baudrate reserved
                          self._parameters['io_voltage'],
                          )
        rv = self._usb.control_transfer_out(
            'device', 'vendor', request=UsbdRequest.EXTIO,
            value=0, index=0, data=msg)
        _ioerror_on_bad_result(rv)

    def _current_ranging_split(self, value):
        if value is None or value in [False, 'off']:
            self.parameter_set('current_ranging_type', 'off')
            return
        parts = value.split('_')
        if len(parts) != 4:
            raise ValueError(f'Invalid current_ranging value {value}')
        for p, v in zip(['type', 'samples_pre', 'samples_window', 'samples_post'], parts):
            self.parameter_set('current_ranging_' + p, v)

    def _current_ranging_set(self):
        if self.stream_buffer is None:
            return
        names = ['current_ranging_type', 'current_ranging_samples_pre',
                 'current_ranging_samples_window', 'current_ranging_samples_post']
        s = '_'.join([str(self.parameter_get(n)) for n in names])
        self.stream_buffer.suppress_mode = s

    def _on_data(self, data):
        # DeviceDriverApi.read_stream_start data_fn callback
        # invoked from USB thread when new sample data is available
        # VERY time critical - keep as short as possible
        # return False to continue streaming, True to stop streaming
        return self.stream_buffer.insert(data)

    def _on_process(self):
        # DeviceDriverApi.read_stream_start process_fn callback
        # invoked from USB thread when new sample data is available after _on_data
        # Time critical, but less so than _on_data
        # return False to continue streaming, True to stop streaming
        rv = False
        try:
            self.stream_buffer.process()
        except Exception:
            log.exception('stream_buffer.process exception: stop streaming')
            return True

        objs = self._process_objs + self._process_objs_add
        self._process_objs = []
        self._process_objs_add = []
        for obj in objs:
            if obj.driver_active:
                try:
                    rv |= bool(obj.stream_notify(self.stream_buffer))
                except Exception:
                    log.exception('%s stream_notify() exception - stop streaming', obj)
                    obj.driver_active = False
                    rv = True
                self._process_objs.append(obj)
            else:
                obj.driver_active = False
                try:
                    if hasattr(obj, 'close'):
                        obj.close()
                except Exception:
                    log.exception('%s close() exception', obj)
        return rv

    def _on_stop(self, status, message):
        # DeviceDriverApi.read_stream_start stop_fn callback
        # invoked from USB thread
        log.info('streaming done(%d, %s)', status, message)
        stop_fn, self._stop_fn = self._stop_fn, None
        if callable(stop_fn):
            stop_fn(status, message)
        for obj in self._process_objs:
            if obj.driver_active and hasattr(obj, 'stop'):
                try:
                    obj.stop()
                except Exception:
                    log.exception('%s stop() exception', obj)

    def _stream_process_call(self, method, *args, **kwargs):
        for obj in self._process_objs:
            fn = getattr(obj, method, None)
            if obj.driver_active and callable(fn):
                try:
                    fn(*args, **kwargs)
                except Exception:
                    log.exception('%s %s() exception', obj, method)
                    obj.driver_active = False

    def start(self, stop_fn=None, duration=None, contiguous_duration=None):
        """Start data streaming.

        :param stop_fn: The function(event, message) called when the device stops.
            The device can stop "automatically" on errors.
            Call :meth:`stop` to stop from the caller.
            This function will be called from the USB processing thread.
            Any calls back into self MUST BE resynchronized.
        :param duration: The duration in seconds for the capture.
        :param contiguous_duration: The contiguous duration in seconds for
            the capture.  As opposed to duration, this ensures that the
            duration has no missing samples.  Missing samples usually
            occur when the device first starts.

        If streaming was already in progress, it will be restarted.
        """
        if self._streaming:
            self.stop()
        self.stream_buffer.reset()

        self._stop_fn = stop_fn
        if duration is not None:
            self.stream_buffer.sample_id_max = int(duration * self._input_sampling_frequency)
        if contiguous_duration is not None:
            c = int(contiguous_duration * self._input_sampling_frequency)
            self.stream_buffer.contiguous_max = c
            log.info('contiguous_samples=%s', c)

        self._process_objs = self._process_objs + self._process_objs_add
        self._stream_process_call('start', stream_buffer=self.stream_buffer)
        self._streaming = True
        self._stream_settings_send()
        self._usb.read_stream_start(
            endpoint_id=2,
            transfers=self._parameters['transfer_outstanding'],
            block_size=self._parameters['transfer_length'] * usb.BULK_IN_LENGTH,
            data_fn=self._on_data,
            process_fn=self._on_process,
            stop_fn=self._on_stop)
        return True

    def stop(self):
        """Stop data streaming.

        :return: True if stopped.  False if was already stopped.

        This method is always safe to call, even after the device has been
        stopped or removed.
        """
        log.info('stop : streaming=%s', self._streaming)
        if self._streaming:
            self._usb.read_stream_stop(2)
            self._streaming = False
            try:
                self._stream_settings_send()
            except Exception:
                log.warning('Device.stop() while attempting _stream_settings_send')
            self._stream_process_call('stop')
            return True
        return False

    def read(self, duration=None, contiguous_duration=None, out_format=None, fields=None):
        """Read data from the device.

        :param duration: The duration in seconds for the capture.
            The duration must fit within the stream_buffer.
        :param contiguous_duration: The contiguous duration in seconds for
            the capture.  As opposed to duration, this ensures that the
            duration has no missing samples.  Missing samples usually
            occur when the device first starts.
            The duration must fit within the stream_buffer.
        :param out_format: The output format which is one of:

            * calibrated: The Nx2 np.ndarray(float32) with columns current and voltage.
            * raw: The raw Nx2 np.ndarray(uint16) Joulescope data.
            * samples_get: The StreamBuffer samples get format.  Use the fields
              parameter to optionally specify the signals to include.
            * None: equivalent to 'calibrated'.

        :param fields: The fields for samples_get when out_format=samples_get.

        If streaming was already in progress, it will be restarted.
        If neither duration or contiguous duration is specified, the capture
        will only be stopped by callbacks registered through
        :meth:`stream_process_register`.
        """
        log.info('read(duration=%s, contiguous_duration=%s, out_format=%s)',
                 duration, contiguous_duration, out_format)
        if out_format not in ['raw', 'calibrated', 'samples_get', None]:
            raise ValueError(f'Invalid out_format {out_format}')
        if duration is None and contiguous_duration is None:
            raise ValueError('Must specify duration or contiguous_duration')
        duration_max = len(self.stream_buffer) / self._output_sampling_frequency
        if contiguous_duration is not None and contiguous_duration > duration_max:
            raise ValueError(f'contiguous_duration {contiguous_duration} > {duration_max} max seconds')
        if duration is not None and duration > duration_max:
            raise ValueError(f'duration {duration} > {duration_max} max seconds')
        q = queue.Queue()

        def on_stop(*args, **kwargs):
            log.info('received stop callback: pending stop')
            q.put(None)

        self.start(on_stop, duration=duration, contiguous_duration=contiguous_duration)
        q.get()
        self.stop()
        start_id, end_id = self.stream_buffer.sample_id_range
        log.info('read available range %s, %s', start_id, end_id)
        if contiguous_duration is not None:
            start_id = end_id - int(contiguous_duration * self._output_sampling_frequency)
        elif duration is not None:
            start_id = end_id - int(duration * self._output_sampling_frequency)
        if start_id < 0:
            start_id = 0
        log.info('read actual %s, %s', start_id, end_id)

        if out_format == 'raw':
            return self.stream_buffer.samples_get(start_id, end_id, fields='raw').reshape((-1, 2))
        elif out_format in ['calibrated', None]:
            return self.stream_buffer.samples_get(start_id, end_id, fields='current_voltage')
        else:
            return self.stream_buffer.samples_get(start_id, end_id, fields=fields)

    @property
    def is_streaming(self):
        """Check if the device is streaming.

        :return: True if streaming.  False if not streaming.
        """
        return self._streaming

    def stream_process_register(self, obj):
        """Register a stream process object.

        :param obj: The instance compatible with :class:`StreamProcessApi`.
            The instance must remain valid until its :meth:`close` is
            called.

        Call :meth:`stream_process_unregister` to disconnect the instance.
        """
        obj.driver_active = True
        self._process_objs_add.append(obj)

    def stream_process_unregister(self, obj):
        """Unregister a stream process object.

        :param obj: The instance compatible with :class:`StreamProcessApi` that was
            previously registered using :meth:`stream_process_register`.
        """
        obj.driver_active = False

    def _status(self):
        def _status_error(ec, msg_str):
            log.warning('status failed %d: %s', ec, msg_str)
            return {'return_code': {'value': ec, 'str': msg_str, 'units': ''}}

        rv = self._usb.control_transfer_in(
            'device', 'vendor',
            request=UsbdRequest.STATUS,
            value=0, index=0, length=STATUS_REQUEST_LENGTH)
        if 0 != rv.result:
            s = 'usb control transfer failed: {}'.format(usb.get_error_str(rv.result))
            return _status_error(rv.result, s)
        pdu = bytes(rv.data)
        v1_expected_length = 8 + 16
        v2_expected_length = 8 + 16 + 4 * 8 + 12 * 4
        expected_lengths = [v1_expected_length, v2_expected_length]
        if len(pdu) not in expected_lengths:
            msg = 'status msg size unexpected: %d not in %s' % (len(pdu), expected_lengths)
            return _status_error(1, msg)
        version, hdr_length, pdu_type = struct.unpack('<BBB', pdu[:3])
        if version != HOST_API_VERSION:
            msg = 'status msg API version mismatch: %d != %d' % (version, HOST_API_VERSION)
            return _status_error(1, msg)
        if pdu_type != PacketType.STATUS:
            msg = 'status msg pdu_type mismatch: %d != %d' % (pdu_type, PacketType.STATUS)
            return _status_error(1, msg)
        if hdr_length != len(pdu):
            msg = 'status msg length mismatch: %d != %d' % (hdr_length, len(pdu))
            return _status_error(1, msg)
        values = struct.unpack('<iIIBBBx', pdu[8:v1_expected_length])
        status = {
            'settings_result': {
                'value': values[0],
                'units': ''},
            'fpga_frame_counter': {
                'value': values[1],
                'units': 'frames'},
            'fpga_discard_counter': {
                'value': values[2],
                'units': 'frames'},
            'sensor_flags': {
                'value': values[3],
                'format': '0x{:02x}',
                'units': ''},
            'sensor_i_range': {
                'value': values[4],
                'format': '0x{:02x}',
                'units': ''},
            'sensor_source': {
                'value': values[5],
                'format': '0x{:02x}',
                'units': ''},
            'return_code': {
                'value': 0,
                'format': '{}',
                'units': '',
            },
        }
        if hdr_length == v2_expected_length:
            s = _statistics_unpack(pdu[v1_expected_length:])
            self._statistics_callback_handler(s)
        for key, value in status.items():
            value['name'] = key

        return status

    def status(self):
        """Get the current device status.

        :return: A dict containing status information.
        """
        status = self._usb.status()
        status['driver'] = self._status()
        status['buffer'] = self.stream_buffer.status()
        return status

    def _sensor_status_check(self):
        rv = self._status()
        ec = rv.get('settings_result', {}).get('value', 1)
        if 0 != ec:
            raise RuntimeError('sensor_firmware_program failed %d' % (ec,))

    def extio_status(self):
        """Read the EXTIO GPI value.

        :return: A dict containing the extio status.  Each key is the status
            item name.  The value is itself a dict with the following keys:

            * name: The status name, which is the same as the top-level key.
            * value: The actual value
            * units: The units, if applicable.
            * format: The recommended formatting string (optional).
        """
        rv = self._usb.control_transfer_in(
            'device', 'vendor',
            request=UsbdRequest.EXTIO,
            value=0, index=0, length=EXTIO_REQUEST_LENGTH)
        if 0 != rv.result:
            s = usb.get_error_str(rv.result)
            log.warning('usb control transfer failed: %s', s)
            return {'return_code': {'value': rv.result, 'str': s, 'units': ''}}
        pdu = bytes(rv.data)
        expected_length = 8 + 16
        if len(pdu) < expected_length:
            log.warning('status msg pdu too small: %d < %d',
                        len(pdu), expected_length)
            return {}
        version, hdr_length, pdu_type = struct.unpack('<BBB', pdu[:3])
        if version != HOST_API_VERSION:
            log.warning('status msg API version mismatch: %d != %d',
                        version, HOST_API_VERSION)
            return {}
        if pdu_type != PacketType.EXTIO:
            return {}
        if hdr_length != expected_length:
            log.warning('status msg length mismatch: %d != %d',
                        hdr_length, expected_length)
            return {}
        values = struct.unpack('<BBBBBBBBII', pdu[8:24])
        status = {
            'flags': {
                'value': values[0],
                'units': ''},
            'trigger_source': {
                'value': values[1],
                'units': ''},
            'current_lsb': {
                'value': values[2],
                'units': ''},
            'voltage_lsb': {
                'value': values[3],
                'units': ''},
            'gpo0': {
                'value': values[4],
                'units': ''},
            'gpo1': {
                'value': values[5],
                'units': ''},
            'gpi_value': {
                'value': values[7],
                'units': '',
            },
            'io_voltage': {
                'value': values[9],
                'units': 'mV',
            },
        }
        for key, value in status.items():
            value['name'] = key
        return status

    def sensor_firmware_program(self, data, progress_cbk=None):
        """Program the sensor microcontroller firmware

        :param data: The firmware to program as a raw binary file.
        :param progress_cbk:  The optional Callable[float] which is called
            with the progress fraction from 0.0 to 1.0
        :raise: on error.
        """
        log.info('sensor_firmware_program')
        if progress_cbk is None:
            progress_cbk = lambda x: None
        self.stop()

        log.info('sensor bootloader: start')
        data = datafile.filename_or_bytes(data)
        if not len(data):
            # erase without programming
            metadata = {
                'size': 0,
                'encryption': 1,
                'header': bytes([0] * 24),
                'mac': bytes([0] * 16),
                'signature': bytes([0] * 64),
            }
        else:
            fh = io.BytesIO(data)
            dr = datafile.DataFileReader(fh)
            # todo: check distribution signature
            tag, hdr_value = next(dr)
            if tag != datafile.TAG_HEADER:
                raise ValueError('incorrect format: expected header, received %r' % tag)
            tag, data = next(dr)
            if tag != datafile.TAG_DATA_BINARY:
                raise ValueError('incorrect format: expected data, received %r' % tag)
            tag, enc = next(dr)
            if tag != datafile.TAG_ENCRYPTION:
                raise ValueError('incorrect format: expected encryption, received %r' % tag)
            metadata = {
                'size': len(data),
                'encryption': 1,
                'header': hdr_value[:24],
                'mac': enc[:16],
                'signature': enc[16:],
            }
        log.info('header    = %r', binascii.hexlify(metadata['header']))
        log.info('mac       = %r', binascii.hexlify(metadata['mac']))
        log.info('signature = %r', binascii.hexlify(metadata['signature']))
        msg = struct.pack('<II', metadata['size'], metadata['encryption'])
        msg = msg + metadata['header'] + metadata['mac'] + metadata['signature']

        rv = self._usb.control_transfer_out(
            'device', 'vendor', request=UsbdRequest.SENSOR_BOOTLOADER,
            value=SensorBootloader.START, data=msg)
        # firmware holds of control transaction complete until done
        _ioerror_on_bad_result(rv)
        self._sensor_status_check()
        time.sleep(1.0)  # give sensor extra time to power up

        log.info('sensor bootloader: erase all flash')
        rv = self._usb.control_transfer_out(
            'device', 'vendor', request=UsbdRequest.SENSOR_BOOTLOADER,
            value=SensorBootloader.ERASE)
        # firmware holds of control transaction complete until done
        _ioerror_on_bad_result(rv)
        self._sensor_status_check()

        log.info('sensor bootloader: program')
        total_size = len(data)
        chunk_size = 2 ** 10  # 16 kB
        assert(0 == (chunk_size % 256))
        index = 0
        while len(data):
            sz = chunk_size if len(data) > chunk_size else len(data)
            fraction_done = (index * 256) / total_size
            progress_cbk(fraction_done)
            log.info('sensor bootloader: program chunk index=%d, sz=%d | %.1f%%', index, sz, fraction_done * 100)
            rv = self._usb.control_transfer_out(
                'device', 'vendor', request=UsbdRequest.SENSOR_BOOTLOADER,
                value=SensorBootloader.WRITE, index=index, data=data[:sz])
            # firmware holds of control transaction complete until done
            _ioerror_on_bad_result(rv)
            self._sensor_status_check()
            data = data[sz:]
            index += chunk_size // 256
        log.info('sensor bootloader: resume')
        rv = self._usb.control_transfer_out(
            'device', 'vendor', request=UsbdRequest.SENSOR_BOOTLOADER,
            value=SensorBootloader.RESUME)
        progress_cbk(1.0)
        _ioerror_on_bad_result(rv)

    def bootloader(self, progress_cbk=None):
        """Start the bootloader for this device.

        :param progress_cbk:  The optional Callable[float] which is called
            with the progress fraction from 0.0 to 1.0
        :return: (bootloader, existing_devices)  Use the bootloader instance
            to perform operations.  Use existing_devices to assist in
            determining when this device returns from bootloader mode.
        """
        if progress_cbk is None:
            progress_cbk = lambda x: None
        _, existing_devices, _ = scan_for_changes(name='Joulescope', devices=[self])
        existing_bootloaders = scan(name='bootloader')
        log.info('my_device = %s', str(self))
        log.info('existing_devices = %s', existing_devices)
        log.info('existing_bootloaders = %s', existing_bootloaders)
        self.stop()
        rv = self._usb.control_transfer_out(
            'device', 'vendor', request=UsbdRequest.CONTROLLER_BOOTLOADER,
            value=SensorBootloader.START)
        _ioerror_on_bad_result(rv)
        self.close()
        b = []
        time_start = time.time()
        while not len(b):
            time_elapsed = time.time() - time_start
            if time_elapsed > USB_RECONNECT_TIMEOUT_SECONDS:
                raise IOError('Timed out waiting for bootloader to connect')
            progress_cbk(time_elapsed / USB_RECONNECT_TIMEOUT_SECONDS)
            time.sleep(0.25)
            _, b, _ = scan_for_changes(name='bootloader', devices=existing_bootloaders)
        if len(b) != 1:
            raise IOError('More than one new bootloader found')
        b = b[0]
        b.open()
        return b, existing_devices

    def run_from_bootloader(self, fn):
        """Run commands from the bootloader and then return to the app.

        :param fn: The function(bootloader) to execute the commands.
        """
        b, existing_devices = self.bootloader()
        try:
            rc = fn(b)
        finally:
            b.go()  # go closes bootloader automatically
            d = []
            time_start = time.time()
            while not len(d):
                if (time.time() - time_start) > USB_RECONNECT_TIMEOUT_SECONDS:
                    raise IOError('Timed out waiting for application to connect')
                time.sleep(0.25)
                _, d, _ = scan_for_changes(name='Joulescope', devices=existing_devices)
            time.sleep(0.5)
            self._usb = d[0]._usb
            self.open()
        return rc

    def controller_firmware_program(self, data, progress_cbk=None):
        return self.run_from_bootloader(lambda b: b.firmware_program(data, progress_cbk))

    def calibration_program(self, data, is_factory=False):
        return self.run_from_bootloader(lambda b: b.calibration_program(data, is_factory))

    def enter_test_mode(self, index=None, value=None):
        """Enter a custom test mode.

        :param index: The test mode index.
        :param value: The test mode value.

        You probably should not be using this method.  You will not destroy
        anything, but you will likely stop your Joulescope from working
        normally until you unplug it.
        """
        index = 0 if index is None else int(index)
        value = 0 if value is None else int(value)
        rv = self._usb.control_transfer_out(
            'device', 'vendor', request=UsbdRequest.TEST_MODE,
            index=index,
            value=value)
        _ioerror_on_bad_result(rv)

    def __enter__(self):
        """Device context manager, automatically open."""
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Device context manager, automatically close."""
        self.close()


def scan(name: str = None, config=None) -> List[Device]:
    """Scan for connected devices.

    :param name: The case-insensitive device name to scan.
        None (default) is equivalent to 'Joulescope'.
    :param config: The configuration for the :class:`Device`.
    :return: The list of :class:`Device` instances.  A new instance is created
        for each detected device.  Use :func:`scan_for_changes` to preserved
        existing instances.
    :raises: None - guaranteed not to raise an exception
    """
    if name is None:
        name = 'joulescope'
    try:
        devices = usb.scan(name)
        if name == 'bootloader':
            devices = [bootloader.Bootloader(d) for d in devices]
        else:
            devices = [Device(d, config=config) for d in devices]
        return devices
    except Exception:
        log.exception('while scanning for devices')
        return []


def scan_require_one(name: str = None, config=None) -> Device:
    """Scan for one and only one device.

    :param name: The case-insensitive device name to scan.
        None (default) is equivalent to 'Joulescope'.
    :param config: The configuration for the :class:`Device`.
    :return: The :class:`Device` found.
    :raise RuntimeError: If no devices or more than one device was found.
    """
    devices = scan(name, config=config)
    if not len(devices):
        raise RuntimeError("no devices found")
    if len(devices) > 1:
        raise RuntimeError("multiple devices found")
    return devices[0]


def scan_for_changes(name: str = None, devices=None, config=None):
    """Scan for device changes.

    :param name: The case-insensitive device name to scan.
        None (default) is equivalent to 'Joulescope'.
    :param devices: The list of existing :class:`Device` instances returned
        by a previous scan.  Pass None or [] if no scan has yet been performed.
    :param config: The configuration for the :class:`Device`.
    :return: The tuple of lists (devices_now, devices_added, devices_removed).
        "devices_now" is the list of all currently connected devices.  If the
        device was in "devices", then return the :class:`Device` instance from
        "devices".
        "devices_added" is the list of connected devices not in "devices".
        "devices_removed" is the list of devices in "devices" but not "devices_now".
    """
    devices_prev = [] if devices is None else devices
    devices_next = scan(name, config=config)
    devices_added = []
    devices_removed = []
    devices_now = []

    for d in devices_next:
        matches = [x for x in devices_prev if str(x) == str(d)]
        if len(matches):
            devices_now.append(matches[0])
        else:
            devices_added.append(d)
            devices_now.append(d)

    for d in devices_prev:
        matches = [x for x in devices_next if str(x) == str(d)]
        if not len(matches):
            devices_removed.append(d)

    log.info('scan_for_changes %d devices: %d added, %d removed',
             len(devices_now), len(devices_added), len(devices_removed))
    return devices_now, devices_added, devices_removed


def bootloaders_run_application():
    """Command all connected bootloaders to run the application.

    :raises: None - guaranteed not to raise an exception.
    """
    log.info('Find all Joulescope bootloaders and run the application')
    for d in scan(name='bootloader'):
        try:
            d.open()
        except Exception:
            log.exception('while attempting to open bootloader')
            continue
        try:
            d.go()
        except Exception:
            log.exception('while attempting to run the application')


def bootloader_go(bootloader, device_name=None, timeout=None, config=None, progress_cbk=None):
    """Command the bootloader to run the application and return the matching device.

    :param bootloader: The target bootloader, which is already open.
    :param device_name: The case-insensitive device name to scan.
        None (default) is equivalent to 'Joulescope'.
    :param timeout: The timeout in seconds while waiting for the application.
    :param config: The configuration for the device.
    :param progress_cbk:  The optional Callable[float] which is called
        with the progress fraction from 0.0 to 1.0
    :return: The matching device, not yet opened.
    :raise IOError: on failure.
    """
    timeout = USB_RECONNECT_TIMEOUT_SECONDS if timeout is None else float(timeout)
    if progress_cbk is None:
        progress_cbk = lambda x: None
    existing_devices = scan(device_name)
    bootloader.go()
    time_start = time.time()
    while True:
        time_elapsed = time.time() - time_start
        if time_elapsed > timeout:
            raise IOError('could not find device')
        progress_cbk(time_elapsed / timeout)
        _, devices, _ = scan_for_changes(device_name, existing_devices, config)
        if len(devices):
            return devices[0]
        time.sleep(0.25)



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from joulescope.driver import scan, scan_require_one, scan_for_changes, \
    bootloaders_run_application, bootloader_go
from joulescope.jls_v2_writer import JlsWriter
import sys
import platform
from .version import __version__, __title__, __description__, __url__, \
    __author__, __author_email__, __license__, __copyright__

VERSION = __version__  # for backwards compatibility
__all__ = ['scan', 'scan_require_one', 'scan_for_changes', 'bootloaders_run_application',
           'bootloader_go', 'JlsWriter',
           '__version__', '__title__', '__description__', '__url__',
           '__author__', '__author_email__', '__license__', '__copyright__']


if sys.hexversion < 0x030700:
    raise RuntimeError('joulescope requires Python 3.7+')


# Although only 64-bit OS/Python is supported, may be able to run on 32bit Python / 32bit Windows.
p_sz, _ = platform.architecture()
is_32bit = sys.maxsize < (1 << 32)

if (is_32bit and '32' not in p_sz) or (not is_32bit and '64' not in p_sz):
    raise RuntimeError('joulescope Python bits must match platform bits')



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module implements readers and writers for tag-length-value files
used by Joulescope.

This file format is used for multiple purposes, including:

* Raw data capture, playback & browsing
* Processed data capture, playback & browsing
* Calibration storage
* Firmware update storage

The file format must meet multiple objectives:

* Support a streaming interface, such as over TCP.
* Support fast access loading from disk.
* Support incremental processing, such as by a microcontroller
  for firmware updates.


The streaming requirement means that seeking back to the start of the
file is not allowed.  Any collections or sections must be indicated with
tags.  However, reading and writing files can seek, so tags may be
rewritten with offset information for improved performance.

The file format starts with a 32 byte header:

* 16 bytes: [0xd3, 0x74, 0x61, 0x67, 0x66, 0x6d, 0x74, 0x20, 0x0d, 0x0a, 0x20, 0x0a, 0x20, 0x20, 0x1a, 0x1c]
* 8 bytes: total length in bytes (0="not provided" or "streaming")
* 3 bytes: reserved (0)
* 1 byte: file version (1)
* 4 bytes: crc32 over this header

The prefix is specially selected to ensure:

* Identification: Help the application determine that
  this file is in the correct format with minimal uncertainty.
* Correct endianness:  Little endian has won, so this entire format is
  stored in little endian format.
* Proper binary processing:  The different line ending combinations
  ensure that the reader is not "fixing" the line endings, since this
  is a binary file format.
* Display: Include "substitute" and "file separator" so that text
  printers to not show the rest of the file.

The remaining file contents are in tag-length-value (TLV) format with CRC32:

* 3 bytes: tag
* 1 byte: TLV flags (compression, encryption)
* 4 bytes: length of data in bytes (may be zero)
* length bytes: The data value
* pad bytes: zero padding to 8 byte + 4 boundary so that crc ends on 8 byte boundary
* 4 bytes: crc32

Tags are selected such that the upper byte is 0.  Since the file format is
little endian, this means that the tag has three usable characters.  The
upper tag bits have the following definitions:

* bit 31: 1=compressed, 0=uncompressed
* bit 30: 1=encrypted, 0=unencrypted, ChaCha20 + Poly1305 with EdDSA signature
* bits [28:24]: reserved
* bits [23:0]: Unique tag

The supported tags include:

* b'HDR': common header information.  Must be the first tag, but
  SGS is optionally allowed before.  Files with encrypted tags will
  typically use the first 24 bytes of this field as the nonce, and then
  increment the last uint32 with each new encrypted block.

  * 8 byte timestamp for data creation.  See time.py for timestamp
    format information.
  * 4 byte version of the file data contents: major8, minor8, patch16.
    If this field is not used, set to 0.
  * 4 byte vendor_id: For USB products the MSB is 0 and LSB is the
    USB VID.
  * 2 byte product_id: unique within vendor_id
  * 2 byte subtype_id: application-defined, unique within product_id.
    A single product may include multiple subtypes, such as
    firmware, FPGA bitstreams and calibration data.  Each product may
    assign values for this field or not use it.
  * 4 byte hardware_compatibility: application-defined.
    Each bit represents a potentially incompatible hardware revision.
    This field should set the bit for each hardware version supported.
    If this field is not used, set to 0.
  * 16 byte serial number identifying device associated with this data.
    If this field is not used, set to 0.

* b'END': Indicate data file end.  Must be the last tag.
* b'CLS': collection start.  The payload is:

  * 8 byte position to the collection end tag.
    This allows fast seeking to skip the collection data.
    In streaming datafile mode, the offset is 0.
  * 2 byte file-specific collection identifier
  * 1 byte collection type: 0=unstructured, 1=list, 2=map
  * 1 byte reserved (0)
  * N bytes: optional application specific data.

* b'CLE': collection end.  May contain application-specific data
  such as indices to increase access performance.
* b'SUB': A subfile, which is often used for storing the calibration
  record inside the data capture.  The payload starts with 128 bytes
  127 bytes of UTF-8 encoded characters) that contains the null-terminated
  file name.  Unused bytes MUST be set to 0.  The remaining payload is
  the file in this datafile format.
* b'IDX': application-specific index information.
* b'MJS': application-specific metadata, JSON formatted.
* b'AJS': application-specific data, JSON formatted.
* b'ABN': application-specific data, binary formatted.
* b'UJS': arbitrary end-user data, JSON formatted.
* b'UBN': arbitrary end-user data, binary formatted.
* b'ENC': encryption authenticity and integrity information.
  This tag must follow every block with the encryption bit set.

  * 16 bytes: ChaCha20 + Poly1305 MAC
  * 64 bytes: EdDSA curve25519 using Blake2b hash (monocypher)
    The signature is computed on the UNENCRYPTED data (sign-then-encrypt)
    For firmware updates, we care more that the firmware is valid than
    who created the cryptotext.  If you want to prevent cryptotext
    forgeries, use encrypt-then-sign with use SGS/SGE and the payload only
    flag.

* b'SGS': signature start.  This field (inclusive) and all others
  up to SGE (exclusive) are included in the signature.  Note that this file
  format makes no provisions for managing keys or ensuring key validity.

  * 1 byte: signature type

    * 1 = EdDSA curve25519 using Blake2b hash (monocypher).

  * 1 byte: flags

    * 1 = include this field (default is exclude)
    * 2 = payload only (exclude tag, length & crc32)

  * 6 bytes: reserved zero
  * 32 bytes: public key

* b'SGE': signature end.  This field is exclude from the signature.
  Payload is the signature.

"""

# Use monocypher crypto library rather than libsodium/NaCl
# Design choice allows for same library on microcontrollers
# http://loup-vaillant.fr/tutorials/chacha20-design
# https://github.com/LoupVaillant/Monocypher/blob/master/src/monocypher.c

import struct
import binascii
from joulescope import time
import zlib
import monocypher
import logging

log = logging.getLogger(__name__)

MAGIC = b'\xd3tagfmt \r\n \n  \x1a\x1c'
assert(len(MAGIC) == 16)
HEADER_SIZE = 32
VERSION = 1

FLAG_COMPRESS = 0x80
FLAG_ENCRYPT = 0x40

TAG_HEADER = b'HDR'  # JSON-formatted header
TAG_END = b'END'
TAG_USB = b'USB'
TAG_COLLECTION_START = b'CLS'
TAG_COLLECTION_END = b'CLE'
TAG_SUBFILE = b'SUB'    # subfile: filename + datafile
TAG_INDEX = b'IDX'
TAG_META_JSON = b'MJS'  # JSON formatted file metadata
TAG_DATA_BINARY = b'ABN'  # binary formatted file data
TAG_DATA_JSON = b'AJS'  # JSON formatted file data
TAG_USER_BINARY = b'UBN'  # binary formatted user data
TAG_USER_JSON = b'UJS'  # JSON formatted user data
TAG_CALIBRATION_JSON = b'CJS'  # deprecated, use TAG_DATA_JSON
TAG_SIGNATURE_START = b'SGS'
TAG_SIGNATURE_END = b'SGE'
TAG_ENCRYPTION = b'ENC'


SIGNATURE_FLAG_KEY_INCLUDE = 1
SIGNATURE_FLAG_PAYLOAD_ONLY = 2


def filename_or_bytes(x):
    if x is None:
        return b''
    elif isinstance(x, str):
        with open(x, 'rb') as f:
            return f.read()
    else:
        return x


def default(value, d):
    if value is None:
        return d
    return value


class Header:
    FORMAT = '<qIIHHI'

    def __init__(self, timestamp=None, version=None,
                 product_id=None, vendor_id=None,
                 subtype_id=None, hardware_compatibility=None,
                 serial_number=None):
        self.timestamp = default(timestamp, time.timestamp_now())
        self.version = default(version, 0)
        self.product_id = default(product_id, 0)
        self.vendor_id = default(vendor_id, 0)
        self.subtype_id = default(subtype_id, 0)
        self.hardware_compatibility = default(hardware_compatibility, 0)
        self.serial_number = default(serial_number, bytes([0] * 16))

    def encode(self):
        v = struct.pack(self.FORMAT,
                        self.timestamp, self.version,
                        self.product_id, self.vendor_id,
                        self.subtype_id, self.hardware_compatibility)
        v += self.serial_number
        return v

    @staticmethod
    def decode(b):
        parts = struct.unpack(Header.FORMAT, b[24:])
        serial_number = b[24:]
        return Header(*parts, serial_number)


class Collection:
    FORMAT = '<QHBB'

    def __init__(self, id_=None, type_=None, end_position=None, start_position=None, data=None):
        self.id_ = 0 if id_ is None else id_
        self.type_ = 0 if type_ is None else type_
        self.end_position = 0 if end_position is None else end_position
        self.start_position = 0 if start_position is None else start_position
        self.data = data
        self.metadata = {}  # application-defined metadata

    def encode(self):
        contents = struct.pack(self.FORMAT, self.end_position, self.id_, self.type_, 0)
        if self.data is not None:
            contents = contents + self.data
        return contents

    @staticmethod
    def decode(data):
        end_position, id_, type_, _ = struct.unpack(Collection.FORMAT, data[:12])
        if len(data) > 12:
            collection_data = data[12:]
        else:
            collection_data = None
        return Collection(id_=id_, type_=type_, end_position=end_position, data=collection_data)

    def __repr__(self):
        parts = ['%s=%r' % (x, getattr(self, x)) for x in ['id_', 'type_', 'end_position', 'start_position']]
        return 'Collection(%s)' % (', '.join(parts), )


def _length_to_pad(length):
    remainder = (length + 4) & 0x7
    if remainder == 0:
        return 0
    else:
        return 8 - remainder


def subfile_split(value):
    """Split a subfile into the name and payload.

    :param value: The value in the SUBFILE tag.
    :return: (name, data).
    """
    name = value.split(b'\x00', 1)[0].decode('utf-8')
    data = value[128:]
    return name, data


def _maybe_compress(data):
    length = len(data)
    flags = 0
    if length:
        dataz = zlib.compress(data, level=1)
        lengthz = len(dataz)
        if lengthz < length:
            data = dataz
            flags = FLAG_COMPRESS
    return flags, data


class DataFileWriter:
    """Create a new instance.

    :param filehandle: The file open for write which must support
        the write, seek and tell methods.
    """

    def __init__(self, filehandle):
        self._fh = filehandle
        self.collections = []  # type: List[Collection]
        self._write_file_header()
        self._signature = None

    def _write_file_header(self):
        length = self._fh.tell()
        self._fh.seek(0)
        header = MAGIC + struct.pack('<QBBBB', length, 0, 0, 0, VERSION)
        crc = binascii.crc32(header)
        header = header + struct.pack('<I', crc)
        assert(HEADER_SIZE == len(header))
        self._fh.write(header)
        if length:
            self._fh.seek(length)

    def _append(self, tag, flags, data):
        """Append a new tag-length-value field to the file.

        :param tag: The 3-byte tag as either an integer or bytes.
        :param flags: The flag bytes value
        :param data: The associated data for the tag (optional).
        :return: The starting position for the tag.
        """
        position = self._fh.tell()
        flags = int(flags)
        if flags < 0 or flags > 255:
            raise ValueError('Invalid flags')
        if isinstance(tag, bytes):
            if len(tag) != 3:
                raise ValueError('Invalid tag length %d' % len(tag))
            tag, = struct.unpack('<I', tag + b'\x00')
        tag &= 0xFFFFFF
        tag |= flags << 24
        length = len(data)
        value = bytearray(8 + length + _length_to_pad(length) + 4)
        value[:8] = struct.pack('<II', tag, length)
        value[8:8 + length] = data
        crc = binascii.crc32(value[:-4])
        value[-4:] = struct.pack('<I', crc)
        self._fh.write(value)
        if self._signature is not None:
            if 0 != (self._signature['flags'] & SIGNATURE_FLAG_PAYLOAD_ONLY):
                self._signature['data'] += data
            else:
                self._signature['data'] += value
        return position

    def append(self, tag, data=None, compress=None):
        """Append a new tag-length-value field to the file.

        :param tag: The 3-byte tag as either an integer or bytes.
        :param data: The associated data for the tag (optional).
        :param compress: When False or None, do not attempt to
            compress the data.  When True, attempt to compress
            the data.
        :return: The starting position for the tag.
        """
        data = b'' if data is None else data
        flags = 0
        if bool(compress):
            flags, data = _maybe_compress(data)
        return self._append(tag, flags, data)

    def append_encrypted(self, tag, data, signing_key, encryption_key, nonce, associated_data=None, compress=None):
        flags = 0
        if bool(compress):
            flags, data = _maybe_compress(data)
        flags |= FLAG_ENCRYPT
        signature = monocypher.signature_sign(signing_key, data)
        mac, data = monocypher.lock(encryption_key, nonce, data, associated_data)
        log.info('signature = %r', binascii.hexlify(signature))
        log.info('mac       = %r', binascii.hexlify(mac))
        pos = self._append(tag, flags, data)
        self._append(TAG_ENCRYPTION, 0, mac + signature)
        return pos

    def append_subfile(self, name: str, data, compress=None):
        """Append a subfile.

        :param name: The name of the subfile, which must fit into 127
            bytes encoded as utf-8.
        :param data: The data in this datafile format.
        :param compress: When False or None, do not attempt to
            compress the data.  When True, attempt to compress
            the data.
        """
        n = name.encode('utf-8')
        if len(n) > 127:
            raise ValueError('name too long: %s' % name)
        payload = n + bytes([0] * (128 - len(n))) + data
        self.append(TAG_SUBFILE, payload, compress)

    def signature_start(self, private_key, flags=None):
        self._signature = {
            'flags': 0 if flags is None else int(flags),
            'key': private_key,
            'data': b'',
        }
        payload = bytearray(8)
        payload[0] = 1  # Ed25519 using Blake2b
        payload[1] = self._signature['flags']
        payload += monocypher.compute_signing_public_key(private_key)
        self.append(TAG_SIGNATURE_START, payload)
        if 0 == (self._signature['flags'] & SIGNATURE_FLAG_KEY_INCLUDE):
            self._signature['data'] = b''

    def signature_end(self):
        s = monocypher.signature_sign(self._signature['key'], self._signature['data'])
        self.append(TAG_SIGNATURE_END, s)

    def append_header(self,
                      timestamp=None, version=None,
                      product_id=None, vendor_id=None,
                      subtype_id=None, hardware_compatibility=None,
                      serial_number=None):
        h = Header(timestamp=timestamp, version=version,
                   product_id=product_id,
                   vendor_id=vendor_id, subtype_id=subtype_id,
                   hardware_compatibility=hardware_compatibility,
                   serial_number=serial_number)
        return self.append(TAG_HEADER, h.encode())

    def collection_start(self, id_=None, type_=None, data=None):
        c = Collection(id_, type_, data=data)
        # note: append writes start_position=0 and end_position=0.
        # Both are later updated by collection_end() in normal operation.
        c.start_position = self.append(TAG_COLLECTION_START, c.encode())
        self.collections.append(c)
        return c

    def collection_end(self, collection=None, data=None):
        c = self.collections.pop()
        if collection is not None and collection != c:
            raise RuntimeError('collection mismatch')
        c.end_position = self.append(TAG_COLLECTION_END, data=data)
        current_position = self._fh.tell()
        self._fh.seek(c.start_position)
        self.append(TAG_COLLECTION_START, c.encode())
        self._fh.seek(current_position)

    def finalize(self):
        for c in self.collections[-1::-1]:
            log.info('collection unclosed: id=%d, type=%d' % (c.id_, c.type_))
            self.collection_end(c)
        self.append(TAG_END)
        self._write_file_header()


def validate_file_header(data):
    if len(data) < HEADER_SIZE:
        raise ValueError('data too short')
    header = data[:HEADER_SIZE]
    if MAGIC != header[:16]:
        raise IOError('invalid file')
    length, version, crc_read = struct.unpack('<QxxxBI', header[16:])
    crc_compute = binascii.crc32(header[:(HEADER_SIZE - 4)])
    if crc_read != crc_compute:
        raise ValueError('invalid header crc: 0x%08x != 0x%08x' % (crc_read, crc_compute))
    if VERSION != version:
        raise ValueError('unsupported file version %d' % version)
    return length, version


def _parse_tag_start(data):
    flags = data[3]
    tag = data[:3]
    length, = struct.unpack('<I', data[4:])
    return tag, flags, length


def _maybe_decompress(value, flags):
    if flags & FLAG_COMPRESS:
        if flags & FLAG_ENCRYPT:
            raise ValueError('Connect decompress encrypted data')
        value = zlib.decompress(value)
    return value


class DataFileReader:
    """Create a new instance.

    :param filehandle: The file-like object open for read.  The file must
        support read, seek and tell.
    """

    def __init__(self, filehandle):
        self._signature_key = None
        self._signature_data = None
        self._fh = filehandle
        header = self._fh.read(HEADER_SIZE)
        if len(header) != HEADER_SIZE:
            raise IOError('file too small')
        self.length, self.version = validate_file_header(header)
        pos1 = self._fh.tell()
        self._fh.seek(0, 2)
        file_length = self._fh.tell()
        self._fh.seek(pos1)
        if self.length == 0:
            log.warning('length was not written, use file length %d', file_length)
            self.length = file_length

    @property
    def remaining_length(self):
        pos = self._fh.tell()
        remaining = self.length - pos
        if remaining < 0:
            remaining = 0
        return remaining

    def __iter__(self):
        return self

    def _read_tag(self):
        remaining_length = self.remaining_length
        if 0 <= remaining_length < 16:
            log.warning('tag read truncated at %d of %d', remaining_length, self._fh.tell())
            return None, 0, b'', b''
        tag_length = self._fh.read(8)
        tag, flags, length = _parse_tag_start(tag_length)
        log.debug('tag read %s length=%d at %d of %d', tag, length, self._fh.tell(), self.length)
        entry = bytearray(8 + length + _length_to_pad(length) + 4)
        entry[:8] = tag_length
        pad = _length_to_pad(length)
        tlv_size = 8 + length + pad + 4
        if remaining_length < tlv_size:
            raise ValueError('File too short')
        entry[8:] = self._fh.read(tlv_size - 8)
        crc = binascii.crc32(entry[:-4])
        crc_read, = struct.unpack('<I', entry[-4:])
        if crc != crc_read:
            log.warning('invalid tag crc: 0x%08x != 0x%08x' % (crc_read, crc))
            # raise ValueError('invalid tag crc: 0x%08x != 0x%08x' % (crc_read, crc))
        value = entry[8:8+length]
        return tag, flags, value, entry

    def tell(self):
        """Give the location of the current entry.

        :return: The position suitable for :meth:`seek`.
        """
        return self._fh.tell()

    def seek(self, position):
        """Change to the location of another entry.

        :param position: The position returned by a previous call
            to :meth:`tell`.
        """
        self._fh.seek(position)
        return position

    def peek_tag_length(self):
        """Peek at the next available entry.

        :return: tuple (tag, value_length)

        This method gets the tag and length quickly.  It does not
        load the data or validate the checksum.
        """
        if 0 <= self.remaining_length < 8:
            return None, 0
        position = self._fh.tell()
        tag_length = self._fh.read(8)
        tag, _, length = _parse_tag_start(tag_length)
        self._fh.seek(position)
        return tag, length

    def peek(self):
        """Peek at the next available entry.

        :return: tuple (tag, value)
        """
        position = self._fh.tell()
        tag, flags, value, _ = self._read_tag()
        self._fh.seek(position)
        value = _maybe_decompress(value, flags)
        return tag, value

    def advance(self):
        """Advance to the next TLV, ignoring data.

        :return: The tag that was skipped.
        """
        if 0 <= self.remaining_length < 16:
            return None
        position = self._fh.tell()
        tag_length = self._fh.read(8)
        tag, _, length = _parse_tag_start(tag_length)
        pad = _length_to_pad(length)
        self._fh.seek(position + 8 + length + pad + 4)
        return tag

    def skip(self):
        """Skip the next available entry, skipping entire collections.

        :return: The tag that was skipped.
        """
        if 0 <= self.remaining_length < 16:
            return None
        position = self._fh.tell()
        tag_length = self._fh.read(8)
        tag, _, length = _parse_tag_start(tag_length)
        if tag == TAG_COLLECTION_START:
            self._fh.seek(position)
            self.collection_goto_end()
            self.skip()
        else:
            pad = _length_to_pad(length)
            self._fh.seek(position + 8 + length + pad + 4)
        return tag

    def collection_goto_end(self):
        """Skip to the collection end.

        :raise RuntimeError: If the current tag is not a COLLECTION_START.
        """
        if 0 <= self.remaining_length < 16:
            return None
        tag, value = next(self)
        if tag != TAG_COLLECTION_START:
            raise RuntimeError('Not a collection start')
        c = Collection.decode(value)
        return self._fh.seek(c.end_position)

    def decrypt(self, signing_key, encryption_key, nonce, associated_data=None):
        """Decrypt the next tag, if needed"""
        tag, flags, value, _ = self._read_tag()
        if flags & FLAG_ENCRYPT:
            tag2, _, value2, _ = self._read_tag()
            if tag2 != TAG_ENCRYPTION:
                raise ValueError('Encrypted data must be followed by ENC tag')
            mac = value2[:16]
            signature = value2[16:]
            value = monocypher.unlock(encryption_key, nonce, mac, value, associated_data)
            if value is None:
                raise ValueError('Decryption failed')
            if not monocypher.signature_check(signature, signing_key, value):
                raise ValueError('Signature check failed')
        if flags & FLAG_COMPRESS:
            value = zlib.decompress(value)
        return tag, value

    def __next__(self):
        """Get the next available entry.

        :return: tuple (tag, value).
        :raise StopIteration: when no more entries remain.
        :raise ValueError: on signature check failure.
        """
        tag, flags, value, entry = self._read_tag()
        if tag is None or tag == TAG_END:
            raise StopIteration()
        if tag == TAG_SIGNATURE_START:
            self._signature_key = bytes(value[8:])
            if value[1] & 1:
                self._signature_data = bytes(entry)
            else:
                self._signature_data = b''
        elif tag == TAG_SIGNATURE_END:
            signature = bytes(value)
            value = monocypher.signature_check(signature, self._signature_key, self._signature_data)
            if not value:
                raise ValueError('signature check failed')
            self._signature_key = None
            self._signature_data = None
        elif self._signature_data is not None:
            self._signature_data += entry
        value = _maybe_decompress(value, flags)
        return tag, value

    def pretty_print(self):
        """Pretty print the datafile structure."""
        indent_str = '  '
        indent = 0
        for tag, value in self:
            if tag == TAG_COLLECTION_END:
                indent -= 1
            tag_str = tag[:3].decode('utf-8')
            if tag == TAG_COLLECTION_START:
                c = Collection.decode(value)
                print("%s%s id=%d, type=%d, start=%d, end=%d" %
                      (indent_str * indent, tag_str, c.id_, c.type_, c.start_position, c.end_position))
                indent += 1
            elif tag == TAG_SIGNATURE_END:
                print("%s%s %s" % (indent_str * indent, tag_str, value))
            else:
                print("%s%s %d" % (indent_str * indent, tag_str, len(value)))



# Copyright 2018 Jetperch LLC.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from joulescope.driver import scan_require_one
from joulescope.firmware_manager import upgrade
import sys


def parser_config(p):
    """Program Joulescope firmware and calibration."""
    p.add_argument('target',
                   choices=['upgrade', 'controller', 'sensor',
                            'calibration_factory', 'calibration_active'],
                   help='The firmware program target')
    p.add_argument('filename',
                   help='The source filename containing the firmware image.')
    return on_run


def _progress(fract):
    # The MIT License (MIT)
    # Copyright (c) 2016 Vladimir Ignatev
    #
    # Permission is hereby granted, free of charge, to any person obtaining
    # a copy of this software and associated documentation files (the "Software"),
    # to deal in the Software without restriction, including without limitation
    # the rights to use, copy, modify, merge, publish, distribute, sublicense,
    # and/or sell copies of the Software, and to permit persons to whom the Software
    # is furnished to do so, subject to the following conditions:
    #
    # The above copyright notice and this permission notice shall be included
    # in all copies or substantial portions of the Software.
    #
    # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
    # INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
    # PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
    # FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT
    # OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
    # OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
    fract = min(max(float(fract), 0.0), 1.0)
    bar_len = 50
    filled_len = int(round(bar_len * fract))
    percents = round(100.0 * fract, 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    sys.stdout.write('[%s] %s%s\r' % (bar, percents, '%'))
    sys.stdout.flush()


def _controller_program_from_bootloader(d, data):
    try:
        d.open()
        rc = d.firmware_program(data)
    finally:
        d.close()
    return rc


def _controller_program_from_app(d, data):
    try:
        d.open()
        rc = d.controller_firmware_program(data)
    finally:
        d.close()
    return rc


def controller_program(data):
    try:
        d = scan_require_one(name='bootloader')
        fn = _controller_program_from_bootloader
    except RuntimeError:
        d = scan_require_one(name='joulescope')
        fn = _controller_program_from_app
    return fn(d, data)


def sensor_program(data):
    d = scan_require_one(name='joulescope')
    try:
        d.open()
        start_time = time.time()
        d.sensor_firmware_program(data, progress_cbk=_progress)
        stop_time = time.time()
    finally:
        d.close()
    return 0


def calibration_program(data, is_factory):
    d = scan_require_one(name='joulescope')
    try:
        d.open()
        start_time = time.time()
        d.calibration_program(data, is_factory=is_factory)
        stop_time = time.time()
    finally:
        d.close()
    return 0


def _upgrade(filename):
    try:
        d = scan_require_one(name='bootloader')
    except RuntimeError:
        d = scan_require_one(name='joulescope')
    d.open()
    upgrade(d, filename, progress_cbk=_progress)
    return 0


def on_run(args):
    if args.filename == '!':
        data = b''
    else:
        with open(args.filename, 'rb') as f:
            data = f.read()
    start_time = time.time()
    if args.target == 'controller':
        rc = controller_program(data)
    elif args.target == 'sensor':
        rc = sensor_program(data)
    elif args.target == 'calibration_factory':
        rc = calibration_program(data, True)
    elif args.target == 'calibration_active':
        rc = calibration_program(data, False)
    elif args.target == 'upgrade':
        rc = _upgrade(args.filename)
    else:
        raise ValueError(f'invalid target: {args.target}')
    stop_time = time.time()
    if rc:
        print('\n***FAILED*** program with error code %r' % rc)
    else:
        print('\nFirmware program took %.2f seconds' % (stop_time - start_time, ))
    return rc




# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Threaded wrapper for the device API.
"""

import threading
import queue
import logging
from joulescope.usb.api import DeviceDriverApi


log = logging.getLogger(__name__)
TIMEOUT = 3.0
TIMEOUT_OPEN = 10.0


def _queue_empty(q):
    while True:
        try:
            q.get(timeout=0.0)
        except queue.Empty:
            break


class DeviceThread:
    """Wrap a :class:`Device` in a thread.

    This class implements the Device API and simply wraps a Device
    implementation so that it runs in its own thread.
    """

    def __init__(self, usb_device: DeviceDriverApi):
        self._device = usb_device
        self._cmd_queue = queue.Queue()  # tuples of (command, args, callback)
        self._signal_queue = queue.Queue()
        self._response_queue = queue.Queue()
        self._thread = None
        self._closing = False
        self._str = None
        self.counter = 0

    def _cmd_process(self, cmd, args, cbk):
        delegate_cbk = False
        rv = None
        try:
            log.debug('_cmd_process %s - start', cmd)
            if cmd == 'status':
                rv = self._device.status()
            elif cmd == 'open':
                event_callback_fn = args
                rv = self._device.open(event_callback_fn)
            elif cmd == 'close':
                rv = self._device.close()
            elif cmd == 'control_transfer_out':
                delegate_cbk = True
                args, kwargs = args
                self._device.control_transfer_out(cbk, *args, **kwargs)
            elif cmd == 'control_transfer_in':
                delegate_cbk = True
                args, kwargs = args
                self._device.control_transfer_in(cbk, *args, **kwargs)
            elif cmd == 'read_stream_start':
                args, kwargs = args
                rv = self._device.read_stream_start(*args, **kwargs)
            elif cmd == 'read_stream_stop':
                args, kwargs = args
                rv = self._device.read_stream_stop(*args, **kwargs)
            elif cmd == '__str__':
                rv = str(self._device)
            else:
                log.warning('unsupported command %s', cmd)
        except Exception:
            log.exception('While running command')
        if not delegate_cbk and callable(cbk):
            try:
                cbk(rv)
            except Exception:
                log.exception('in callback')
        log.debug('_cmd_process %s - done', cmd)

    def _cmd_process_all(self):
        _quit = False
        try:
            while not _quit:
                cmd, args, cbk = self._cmd_queue.get(timeout=0.0)
                self.counter += 1
                self._cmd_process(cmd, args, cbk)
                if cmd in ['close']:
                    log.info('DeviceThread._cmd_process_all close')
                    _quit = True
        except queue.Empty:
            pass
        except Exception:
            log.exception('DeviceThread._cmd_process_all unhandled')
        return _quit

    def _cmd_flush(self):
        while True:
            try:
                cmd, args, cbk = self._cmd_queue.get(timeout=0.0)
                self.counter += 1
                if not callable(cbk):
                    continue
                cbk(ConnectionError('device closed'))
            except queue.Empty:
                break
            except Exception:
                log.exception('_cmd_flush')

    def run(self):
        _quit = False
        log.info('DeviceThread.run start')
        while not _quit:
            try:
                self._device.process(timeout=0.05)
            except Exception:
                log.exception('In device thread')
            _quit = self._cmd_process_all()
        log.info('DeviceThread.run flush')
        self._cmd_flush()
        log.info('DeviceThread.run done')

    def _post(self, command, args, cbk):
        # log.debug('DeviceThread %s', command)
        if self._thread is None:
            log.info('DeviceThread.post(%s) when thread not running', command)
        else:
            self._cmd_queue.put((command, args, cbk))
            self._device.signal()

    def _join(self, timeout=None):
        timeout = TIMEOUT if timeout is None else timeout
        if not self._closing:
            self._closing = True
            self._post('close', None, None)
        if self._thread:
            # thread can safely join() multiple times
            self._thread.join(timeout=timeout)
            self._thread = None

    def _post_block(self, command, args, timeout=None):
        timeout = TIMEOUT if timeout is None else float(timeout)
        log.debug('_post_block %s start', command)
        while not self._response_queue.empty():
            log.warning('response queue not empty')
            try:
                self._response_queue.get(timeout=0.0)
            except queue.Empty:
                pass
        self._post(command, args, lambda rv_=None: self._response_queue.put(rv_))
        if self._thread is None:
            raise IOError('DeviceThread not running')
        else:
            try:
                rv = self._response_queue.get(timeout=timeout)
            except queue.Empty as ex:
                log.error('device thread hung: %s - FORCE CLOSE', command)
                self._join(timeout=TIMEOUT)
                rv = ex
            except Exception as ex:
                rv = ex
        if isinstance(rv, Exception):
            raise IOError(rv)
        log.debug('_post_block %s done', command)  # rv
        return rv

    def __str__(self):
        if self._str is not None:
            pass
        elif self._thread is not None:
            self._str = self._post_block('__str__', None)
        else:
            self._str = str(self._device)
        return self._str

    @property
    def serial_number(self):
        if self._device is None:
            return None
        return self._device.serial_number

    def open(self, event_callback_fn=None):
        self.close()
        log.info('open')
        self._thread = threading.Thread(name='usb_device', target=self.run)
        self._thread.start()
        self._closing = False
        try:
            return self._post_block('open', event_callback_fn, timeout=TIMEOUT_OPEN)
        except Exception:
            self.close()
            raise

    def close(self):
        log.info('close')
        self._join(timeout=TIMEOUT)
        _queue_empty(self._cmd_queue)
        _queue_empty(self._response_queue)
        _queue_empty(self._signal_queue)

    def control_transfer_out(self, *args, **kwargs):
        return self._post_block('control_transfer_out', (args, kwargs))

    def control_transfer_in(self, *args, **kwargs):
        return self._post_block('control_transfer_in', (args, kwargs))

    def read_stream_start(self, *args, **kwargs):
        return self._post_block('read_stream_start', (args, kwargs))

    def read_stream_stop(self, *args, **kwargs):
        return self._post_block('read_stream_stop', (args, kwargs))

    def status(self):
        return self._post_block('status', None)

    def signal(self):
        self._signal_queue.put(None)

    def process(self, timeout=None):
        try:
            self._signal_queue.get(timeout=timeout)
        except queue.Empty:
            pass



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This executable captures the raw USB stream from Joulescope devices
and saves the raw stream data to a file.  This executable is a
development tool and is not intended for customer use.
"""

import signal
import time
from joulescope.pattern_buffer import PatternBuffer
import logging
from joulescope import scan


def parser_config(p):
    """Capture raw USB data from Joulescope."""
    p.add_argument('--duration',
                   type=float,
                   help='The capture duration.')
    p.add_argument('--filename',
                   help='The filename for output data.')
    return on_cmd


def on_cmd(args):
    d = scan(name='Joulescope')
    if not(len(d)):
        print('No devices found')
        return 1
    elif len(d) != 1:
        print('More than on device found)')
        return 2
    device = d[0]
    return run(device, filename=args.filename, duration=args.duration)


class RecordingBuffer:

    def __init__(self, filename):
        self.fh = open(filename, 'wb')
        self._sample_id_count = 0
        self.sample_id_max = None
        self.contiguous_max = None

    def close(self):
        if self.fh is not None:
            self.fh.close()
            self.fh is None

    def status(self):
        return {}

    def reset(self):
        self._sample_id_count = 0

    def calibration_set(self, *args, **kwargs):
        pass

    def insert(self, data):
        if self.sample_id_max is not None and self._sample_id_count >= self.sample_id_max or data is None:
            if self.fh is not None:
                self.fh.close()
            return True
        elif self.fh is not None:
            self.fh.write(data)
        self._sample_id_count += (len(data) // 512) * 126
        return False

    def process(self):
        return False


def run(device, filename, duration):
    logging.basicConfig(level=logging.DEBUG)
    time_last = time.time()
    quit = False
    if filename:
        buffer = RecordingBuffer(filename=filename)
    else:
        buffer = PatternBuffer()

    def do_quit(*args, **kwargs):
        nonlocal quit
        quit = 'quit from SIGINT'

    def do_stop(*args, **kwargs):
        nonlocal quit
        quit = 'quit from done'

    signal.signal(signal.SIGINT, do_quit)
    device.open()
    try:
        device.stream_buffer = buffer
        device.parameter_set('sensor_power', 'on')
        # device.parameter_set('control_test_mode', 'normal')
        device.parameter_set('source', 'pattern_sensor')
        device.start(stop_fn=do_stop, duration=duration)
        print('Press CTRL-C to stop data collection')
        while not quit:
            time.sleep(0.01)
            time_now = time.time()
            if time_now - time_last > 1.0:
                print(device.status())
                time_last = time_now
        print(device.status())
    finally:
        device.close()
    print('done capturing data: %s' % quit)

    return 0



# Copyright 2020 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the array_starage
"""

import unittest
from joulescope import array_storage
import numpy as np


class TestArrayStorage(unittest.TestCase):

    def test_empty(self):
        b = array_storage.pack({}, sample_count=0)
        y, c = array_storage.unpack(b)
        self.assertEqual(c, 0)
        self.assertEqual(len(y), 0)

    def test_single_u8(self):
        x = np.arange(256, dtype=np.uint8)
        b = array_storage.pack({2: x}, sample_count=42)
        y, c = array_storage.unpack(b)
        self.assertEqual(c, 42)
        self.assertIn(2, y)
        np.testing.assert_equal(y[2], x)

    def test_invalid_array_id(self):
        x = np.arange(256, dtype=np.uint8)
        with self.assertRaises(ValueError):
            array_storage.pack({-1: x}, sample_count=42)

    def test_invalid_sample_count(self):
        with self.assertRaises(ValueError):
            array_storage.pack({}, sample_count=-1)
        array_storage.pack({}, sample_count=0)
        array_storage.pack({}, sample_count=2 ** 24 - 1)
        with self.assertRaises(ValueError):
            array_storage.pack({}, sample_count=2**24)

    def test_single_float32(self):
        x = np.arange(0, 10, 0.25, dtype=np.float32)
        b = array_storage.pack({3: x}, sample_count=43)
        y, c = array_storage.unpack(b)
        self.assertEqual(c, 43)
        self.assertIn(3, y)
        np.testing.assert_equal(y[3], x)

    def test_multiple(self):
        x_u8 = np.arange(256, dtype=np.uint8)
        x_f32 = np.arange(0, 10, 0.25, dtype=np.float32)
        x = {9: x_u8, 5: x_f32}
        b = array_storage.pack(x, sample_count=44)
        y, c = array_storage.unpack(b)
        self.assertEqual(c, 44)
        self.assertEqual(len(y), 2)
        np.testing.assert_equal(y[9], x_u8)
        np.testing.assert_equal(y[5], x_f32)

    def test_unpack_empty(self):
        with self.assertRaises(ValueError):
            array_storage.unpack(np.zeros(0, dtype=np.uint8))

    def test_unpack_invalid(self):
        with self.assertRaises(ValueError):
            array_storage.unpack(np.zeros(100, dtype=np.uint8))

    def test_truncated(self):
        x = np.arange(0, 10, 0.25, dtype=np.float32)
        b = array_storage.pack({3: x}, sample_count=43)
        with self.assertRaises(ValueError):
            array_storage.unpack(b[:-16])
